"""
File system helpers
"""

import functools
import os
import pathlib
import re
import time

from .. import __project_name__, constants, errors
from . import LazyModule, os_family, types

import logging  # isort:skip
_log = logging.getLogger(__name__)

natsort = LazyModule(module='natsort', namespace=globals())


OS_SEP_REGEX = re.escape(os.sep)
r"""
:attr:`os.sep` escaped with :func:`re.escape`

On Windows, this should be ``"\\"`` while it should be identical to
:attr:`os.sep` on Unix-likes.
"""


def assert_file_readable(filepath):
    """Raise :exc:`~.ContentError` if `open(filepath)` fails"""
    try:
        open(filepath).close()
    except OSError as e:
        raise errors.ContentError(f'{filepath}: {e.strerror or e}')


def assert_dir_usable(path):
    """Raise :exc:`~.ContentError` if `dirpath` can not be used as a directory"""
    if not os.path.isdir(path):
        raise errors.ContentError(f'{path}: Not a directory')
    elif not os.access(path, os.R_OK):
        raise errors.ContentError(f'{path}: Not readable')
    elif not os.access(path, os.W_OK):
        raise errors.ContentError(f'{path}: Not writable')
    elif not os.access(path, os.X_OK):
        raise errors.ContentError(f'{path}: Not executable')


@functools.lru_cache(maxsize=None)
def projectdir(content_path, base=None):
    """
    Return path to existing directory in which jobs put their files and cache

    :param str content_path: Path to torrent content
    :param str base: Location of the project directory; defaults to
        :attr:`~.constants.DEFAULT_CACHE_DIRECTORY`

    :raise ContentError: if `content_path` exists and is not a directory or has
        insufficient permissions
    """
    if content_path:
        if not base:
            base = constants.DEFAULT_CACHE_DIRECTORY
        path = os.path.join(base, basename(content_path) + f'.{__project_name__}')
        mkdir(path)
    else:
        path = '.'
    _log.debug('Using project directory: %r', path)
    return path


def limit_directory_size(path, max_total_size, min_age=None, max_age=None):
    """
    Delete oldest files (by access time) until maximum size is not exceeded

    Empty files and directories are always deleted.

    :param path: Path to directory
    :param max_total_size: Maximum combined size of all files in `path` and its
        subdirectories
    :param min_age: Preserve files that are younger than this
    :type min_age: Unix timestamp
    :param max_age: Preserve files that are older than this
    :type max_age: Unix timestamp
    """
    @functools.lru_cache(maxsize=None)
    def cached_file_size(f):
        return file_size(f)

    def combined_size(filepaths):
        return sum(cached_file_size(f) for f in filepaths
                   if os.path.exists(f) and not os.path.islink(f))

    # This should return mtime if file system was mounted with noatime.
    def atime(filepath):
        try:
            return os.stat(filepath, follow_symlinks=False).st_atime
        except OSError:
            return 0

    def get_filepaths(dirpath):
        return file_list(dirpath, min_age=min_age, max_age=max_age, follow_dirlinks=False)

    # How much space do we have to free up?
    filepaths = get_filepaths(path)
    total_size = combined_size(filepaths)
    size_diff = total_size - max_total_size
    if size_diff > 0:
        # Collect old files until they free up enough space when removed
        oldest_files = sorted(filepaths, key=atime)
        files_to_remove = [oldest_files.pop(0)]
        while combined_size(files_to_remove) < size_diff:
            files_to_remove.append(oldest_files.pop(0))

        # Actually remove the files
        for file_to_remove in files_to_remove:
            # It's possible another process already unlinked this file while we
            # were still collecting old files.
            try:
                os.unlink(file_to_remove)
            except FileNotFoundError:
                pass


def prune_empty(path, files=False, directories=True):
    """
    Remove empty subdirectories recursively

    :param path: Path to directory
    :param bool files: Whether to prune empty files
    :param bool directories: Whether to prune empty directories

    Dead symbolic links are removed after pruning files and directories.

    If `path` is not a directory, do nothing.
    """
    if not os.path.isdir(path):
        return

    def raise_error(e, path):
        raise RuntimeError(f'{path}: Failed to prune: {e.strerror or e}')

    # Prune empty files
    if files:
        for dirpath, dirnames, filenames in os.walk(path, topdown=False, followlinks=False):
            for filename in filenames:
                filepath = os.path.join(dirpath, filename)
                try:
                    if os.path.exists(filepath) and file_size(filepath) <= 0:
                        os.unlink(filepath)
                except OSError as e:
                    raise_error(e, filepath)

    # Prune empty directories
    if directories:
        for dirpath, dirnames, filenames in os.walk(path, topdown=False, followlinks=False):
            for dirname in dirnames:
                subdirpath = os.path.join(dirpath, dirname)
                try:
                    if not os.path.islink(subdirpath) and not os.listdir(subdirpath):
                        os.rmdir(subdirpath)
                except OSError as e:
                    raise_error(e, subdirpath)

    # Prune dead links
    for dirpath, dirnames, filenames in os.walk(path, topdown=False, followlinks=False):
        for filename in filenames:
            filepath = os.path.join(dirpath, filename)
            try:
                if os.path.islink(filepath):
                    if not os.path.exists(filepath):
                        os.unlink(filepath)
            except OSError as e:
                raise_error(e, filepath)

    try:
        if directories and not os.listdir(path):
            os.rmdir(path)
    except OSError as e:
        raise_error(e, path)


def mkdir(path):
    """
    Create directory and its parents

    Existing directories are ignored.

    If `path` is empty, no directory is created.

    :raise ContentError: if directory creation fails
    """
    if path:
        try:
            os.makedirs(path, exist_ok=True)
        except OSError as e:
            raise errors.ContentError(f'{path}: {e.strerror or e}')
        else:
            assert_dir_usable(path)


def basename(path, parents=0):
    """
    Return last segment in `path`

    Unlike :func:`os.path.basename`, this removes any trailing directory
    separators first.

    >>> os.path.basename('a/b/c/')
    ''
    >>> upsies.utils.basename('a/b/c/')
    'c'

    :param parents: How many levels of parent directories to include in the
        returned path
    """
    basename = os.path.basename(str(path).rstrip(os.sep))
    parent_parts = list(pathlib.Path(str(path)).parent.parts)
    parents_to_include = []
    for _ in range(parents):
        if parent_parts:
            parents_to_include.insert(0, parent_parts.pop(-1))
    return str(pathlib.Path(*parents_to_include).joinpath(basename))


def dirname(path):
    """
    Remove last segment in `path`

    Unlike :func:`os.path.dirname`, this removes any trailing directory
    separators first.

    >>> os.path.dirname('a/b/c/')
    'a/b/c'
    >>> upsies.utils.dirname('a/b/c/')
    'a/b'
    """
    return os.path.dirname(str(path).rstrip(os.sep))


def file_and_parent(path):
    """Return `path`'s filename and parent directoy name if there is one"""
    if os.sep in path:
        return (
            basename(path),
            basename(dirname(path)),
        )
    else:
        return (path,)


def sanitize_filename(filename):
    """
    Replace illegal characters in `filename` with "_"

    Illegal characters include :attr:`os.sep`.
    """
    if os_family() == 'windows':
        illegal_chars = ('<', '>', ':', '"', '/', '\\', '|', '?', '*')
    else:
        illegal_chars = ('/',)
    for char in illegal_chars:
        filename = filename.replace(char, '_')
    return filename


def sanitize_path(path):
    """
    Replace illegal characters in each `path` segment with "_"

    `path` is split at :attr:`os.sep` and the resulting items are passed to
    :func:`sanitize_filename` and joined.

    On Windows, the drive (e.g. "C:") is not sanitized to keep the ":".
    """
    # Get drive from Windows paths (e.g. "C:") because we don't want to sanitize
    # the ":", which is illegal otherwise
    drive, path = os.path.splitdrive(str(path))
    segments = str(path).split(os.sep)
    sanitized_segments = [sanitize_filename(segment) for segment in segments]
    return drive + os.sep.join(sanitized_segments)


def tildify_path(path):
    """Return `path` with $HOME replaced by ``~``"""
    home = os.path.expanduser('~')
    return re.sub(rf'^{re.escape(home)}', '~', path)


def file_extension(path, minlen=1, maxlen=None, only=()):
    """
    Return file extension

    Unlike :func:`os.path.splitext`, this function expects the extension to
    consist only of alphanumeric ASCII characters.

    :param str path: Path to file or directory
    :param str minlen: Minimum number of characters in file extension or `None`
        for no limit
    :param str maxlen: Maximum number of characters in file extension or `None`
        for no limit
    :param only: Only strip extension if it exists in this sequence
    :type only: sequence of :class:`str`

    :return: file extension
    :rtype: str
    """
    filename = os.path.basename(path)
    match = re.search(rf'\.([a-zA-Z0-9]{_make_minmax(minlen, maxlen)})$', filename)
    if match:
        ext = match.group(1).lower()
        if not only or ext in (str(e).lower() for e in only):
            return ext
    return ''

def strip_extension(path, minlen=1, maxlen=None, only=()):
    """
    Return `path` without file extension

    If `path` doesn't have a file extension, return it as is.

    A file extension consists only of alphanumeric ASCII characters.

    :param str path: Path to file or directory
    :param str minlen: Minimum number of characters in file extension or `None`
        for no limit
    :param str maxlen: Maximum number of characters in file extension or `None`
        for no limit
    :param only: Only strip extension if it exists in this sequence
    :type only: sequence of :class:`str`

    :return: file name without extension
    :rtype: str
    """
    path = str(path)
    match = re.search(rf'^(.*)\.([a-zA-Z0-9]{_make_minmax(minlen, maxlen)})$', path)
    if match:
        name = match.group(1)
        ext = match.group(2).lower()
        if not only or ext in (str(e).lower() for e in only):
            return name
    return path

def _make_minmax(minlen, maxlen):
    if minlen is not None and maxlen is not None:
        return '{' + f'{minlen},{maxlen}' + '}'
    elif minlen is not None:
        return '{' + f'{minlen},' + '}'
    elif maxlen is not None:
        return '{' + f',{maxlen}' + '}'
    else:
        return '*'


def file_size(path):
    """Return file size in bytes or `None`"""
    if not os.path.isdir(path):
        try:
            return os.path.getsize(path)
        except OSError:
            pass
    return None


def path_size(path):
    """
    Combined size of all files beneath `path`

    Returns the same value as :func:`file_size` if `path` is not a directory.

    Symbolic links are ignored.
    """
    filepaths = file_list(path, follow_dirlinks=False)
    return sum(
        file_size(f) or 0
        for f in filepaths
        if not os.path.islink(f)
    )


def file_list(path, extensions=(), min_age=None, max_age=None, follow_dirlinks=False):
    """
    List naturally sorted files in `path` and any subdirectories

    If `path` is not a directory, it is returned as a single item in a list
    unless `extensions` are given and they don't match.

    If `path` is any falsy value (e.g. ``""`` or `None`), return an empty
    sequence.

    Unreadable directories are excluded.

    :param str path: Path to a directory
    :param str extensions: Only include files with one of these extensions
        (matched case-insensitively)
    :param min_age: Exclude files that are younger than this
    :type min_age: int or float
    :param max_age: Exclude files that are older than this
    :type max_age: int or float
    :param follow_dirlinks: Whether to include the contents of symbolic links to
        directories or the links themselves

    :return: Tuple of file paths
    """
    extensions = tuple(str(e).casefold() for e in extensions)

    def ext_ok(filename):
        return not extensions or file_extension(filename).casefold() in extensions

    def age_ok(filepath, now=time.time()):
        try:
            statinfo = os.stat(filepath, follow_symlinks=False)
        except OSError:
            return False
        else:
            age = round(now - statinfo.st_atime)
            return (min_age or age) <= age <= (max_age or age)

    if not path:
        # Handle any false value, e.g. "" or `None`
        return ()
    elif not os.path.isdir(path):
        if ext_ok(path):
            return (str(path),)
        else:
            return ()

    files = []
    for root, dirnames, filenames in os.walk(path, topdown=True, followlinks=follow_dirlinks):
        for filename in sorted(filenames):
            filepath = os.path.join(root, filename)
            if ext_ok(filename) and age_ok(filepath):
                files.append(filepath)

        # Symbolic links to directories are in `dirnames`
        if not follow_dirlinks:
            # For symbolic links that point to a directory, os.walk() can only
            # include the contents (followlinks=True) or completely ignore the
            # link. Here we look for links to directories and a) include them in
            # the returned list and b), remove them from `dirnames` so os.walk()
            # doesn't recurse into them. It is important to pass topdown=True
            # for b) to work.
            for dirname in tuple(dirnames):
                # Contrary to what the docs say, this returns `True` for dead
                # and living symbolic links
                dirpath = os.path.join(root, dirname)
                if os.path.islink(dirpath):
                    del dirnames[dirnames.index(dirname)]
                    files.append(dirpath)

    return tuple(natsort.natsorted(files, key=str.casefold))


def find_name(name, path, validator=None):
    """
    Return first path (in human sort order) to `name` anywhere beneath
    `path` or `None`

    :param str name: Exact name of the file or directory
    :param str path: Base path to search for `name`

    :param validator: Callable that gets a path that ends with `name` and
        returns whether it is a match or not

        For example, you can use :func:`os.path.isfile` or :func:`os.path.isdir`
        to only find files or directories.
    """
    for root_, dirnames_, filenames_ in natsort.natsorted(
            os.walk(path, topdown=True, followlinks=False),
    ):
        names_ = natsort.natsorted(filenames_ + dirnames_)
        for name_ in names_:
            path_ = os.path.join(root_, name_)
            if name_ == name and (
                    not validator or validator(path_)
            ):
                return path_


# Stolen from torf-cli
# https://github.com/rndusr/torf-cli/blob/master/torfcli/_utils.py

_DOWN       = '\u2502'  # noqa: E221  # │
_DOWN_RIGHT = '\u251C'  # noqa: E221  # ├
_RIGHT      = '\u2500'  # noqa: E221  # ─
_CORNER     = '\u2514'  # noqa: E221  # └

def format_file_tree(tree, _parents_is_last=()):
    """
    Format nested file tree sequence as indented multi-line string

    :param tree: Nested 2-tuples: The first item is the file or directory name,
        the second item is the file size for files or a tuple for directories
    """
    lines = []
    max_i = len(tree) - 1
    for i,(name,node) in enumerate(tree):
        is_last = i >= max_i

        # Assemble indentation string (`_parents_is_last` being empty means
        # this is the top node)
        indent = ''
        if _parents_is_last:
            # `_parents_is_last` holds the `is_last` values of our ancestors.
            # This lets us construct the correct indentation string: For
            # each parent, if it has any siblings below it in the directory,
            # print a vertical bar ('|') that leads to the siblings.
            # Otherwise the indentation string for that parent is empty.
            # We ignore the first/top/root node because it isn't indented.
            for parent_is_last in _parents_is_last[1:]:
                if parent_is_last:
                    indent += '  '
                else:
                    indent += f'{_DOWN} '

            # If this is the last node, use '└' to stop the line, otherwise
            # branch off with '├'.
            if is_last:
                indent += f'{_CORNER}{_RIGHT}'
            else:
                indent += f'{_DOWN_RIGHT}{_RIGHT}'

        if isinstance(node, int):
            lines.append(f'{indent}{name} ({types.Bytes(node)})')
        else:
            def sum_node_size(nodelist):
                size = 0
                for node in nodelist:
                    if isinstance(node, tuple):
                        if len(node) >= 2 and isinstance(node[1], int):
                            size += node[1]
                        else:
                            size += sum_node_size(node)
                return size

            lines.append(f'{indent}{name} ({types.Bytes(sum_node_size(node))})')
            # Descend into child node
            sub_parents_is_last = _parents_is_last + (is_last,)
            lines.append(format_file_tree(node, _parents_is_last=sub_parents_is_last))

    return '\n'.join(lines)



import re
from unittest.mock import Mock, call

import pytest

from upsies import errors
from upsies.jobs.mediainfo import MediainfoJob


@pytest.fixture
def job(tmp_path):
    return MediainfoJob(
        home_directory=tmp_path,
        cache_directory=tmp_path,
        content_path='path/to/content',
    )


@pytest.mark.parametrize(
    argnames='from_all_videos, exp_from_all_videos_id',
    argvalues=(
        (True, 'from_all_videos'),
        (False, 'from_first_video'),
    ),
)
@pytest.mark.parametrize(
    argnames='exclude_files, exp_exclude_files_id',
    argvalues=(
        ((), []),
        (('*.foo', 'bar*'), ['*.foo', 'bar*']),
        ((re.compile(r'.*\.foo$'), re.compile(r'^bar.*')), [r'.*\.foo$', '^bar.*']),
        (('*.foo', re.compile(r'^bar.*')), ['*.foo', '^bar.*']),
    ),
    ids=lambda v: repr(v),
)
def test_MediainfoJob_cache_id(from_all_videos, exp_from_all_videos_id,
                               exclude_files, exp_exclude_files_id,
                               job, mocker):
    mocker.patch.object(job, '_content_path', 'path/to/Foo.2000-ASDF')
    mocker.patch.object(job, '_from_all_videos', from_all_videos)
    mocker.patch.object(job, '_exclude_files', exclude_files)
    assert job.cache_id == ('Foo.2000-ASDF', exp_from_all_videos_id, exp_exclude_files_id)


def test_MediainfoJob_initialize():
    job = MediainfoJob(
        content_path='path/to/my/content',
        from_all_videos='maybe?',
        exclude_files=('foo', 'bar', 'baz'),
    )
    assert job._content_path == 'path/to/my/content'
    assert job._from_all_videos == 'maybe?'
    assert job._exclude_files == ('foo', 'bar', 'baz')
    assert job.signal.signals['mediainfo'] == [job._store_mediainfos_by_file]


def test_MediainfoJob_execute(job, mocker):
    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(job, 'attach_task'), 'attach_task')
    mocks.attach_mock(mocker.patch.object(job, '_generate_mediainfos', Mock()), '_generate_mediainfos')

    job.execute()

    assert mocks.mock_calls == [
        call._generate_mediainfos(),
        call.attach_task(
            coro=job._generate_mediainfos.return_value,
            finish_when_done=True,
        ),
    ]


@pytest.mark.parametrize(
    argnames='filtered_videos, from_all_videos, mediainfo_exceptions, exp_mock_calls',
    argvalues=(
        pytest.param(
            ('foo.mkv', 'bar.mkv', 'baz.mkv'),
            False,
            {
                'foo.mkv': None,
                'bar.mkv': None,
                'baz.mkv': None,
            },
            [
                call.mediainfo('foo.mkv'),
                call.send('<mediainfo for foo.mkv>'),
                call.emit('mediainfo', 'foo.mkv', '<mediainfo for foo.mkv>'),
            ],
            id='MKV / First / No exception',
        ),

        pytest.param(
            ('foo.mkv', 'bar.mkv', 'baz.mkv'),
            False,
            {
                'foo.mkv': None,
                'bar.mkv': errors.ContentError('no'),
                'baz.mkv': None,
            },
            [
                call.mediainfo('foo.mkv'),
                call.send('<mediainfo for foo.mkv>'),
                call.emit('mediainfo', 'foo.mkv', '<mediainfo for foo.mkv>'),
            ],
            id='MKV / First / ContentError is irrelevant',
        ),

        pytest.param(
            ('foo.mkv', 'bar.mkv', 'baz.mkv'),
            False,
            {
                'foo.mkv': errors.ContentError('no'),
                'bar.mkv': None,
                'baz.mkv': None,
            },
            [
                call.mediainfo('foo.mkv'),
                call.error(errors.ContentError('no')),
            ],
            id='MKV / First / ContentError is reported',
        ),

        pytest.param(
            ('foo.mkv', 'bar.mkv', 'baz.mkv'),
            False,
            {
                'foo.mkv': None,
                'bar.mkv': Exception('NO'),
                'baz.mkv': None,
            },
            [
                call.mediainfo('foo.mkv'),
                call.send('<mediainfo for foo.mkv>'),
                call.emit('mediainfo', 'foo.mkv', '<mediainfo for foo.mkv>'),
            ],
            id='MKV / First / Exception is irrelevant',
        ),

        pytest.param(
            ('foo.mkv', 'bar.mkv', 'baz.mkv'),
            False,
            {
                'foo.mkv': Exception('NO'),
                'bar.mkv': None,
                'baz.mkv': None,
            },
            [
                call.mediainfo('foo.mkv'),
                Exception('NO'),
            ],
            id='MKV / First / Exception is raised',
        ),

        pytest.param(
            ('foo.mkv', 'bar.mkv', 'baz.mkv'),
            True,
            {
                'foo.mkv': None,
                'bar.mkv': None,
                'baz.mkv': None,
            },
            [
                call.mediainfo('foo.mkv'),
                call.send('<mediainfo for foo.mkv>'),
                call.emit('mediainfo', 'foo.mkv', '<mediainfo for foo.mkv>'),
                call.mediainfo('bar.mkv'),
                call.send('<mediainfo for bar.mkv>'),
                call.emit('mediainfo', 'bar.mkv', '<mediainfo for bar.mkv>'),
                call.mediainfo('baz.mkv'),
                call.send('<mediainfo for baz.mkv>'),
                call.emit('mediainfo', 'baz.mkv', '<mediainfo for baz.mkv>'),
            ],
            id='MKV / All / No exception',
        ),

        pytest.param(
            ('foo.mkv', 'bar.mkv', 'baz.mkv'),
            True,
            {
                'foo.mkv': None,
                'bar.mkv': errors.ContentError('no'),
                'baz.mkv': None,
            },
            [
                call.mediainfo('foo.mkv'),
                call.send('<mediainfo for foo.mkv>'),
                call.emit('mediainfo', 'foo.mkv', '<mediainfo for foo.mkv>'),
                call.mediainfo('bar.mkv'),
                call.error(errors.ContentError('no')),
                call.mediainfo('baz.mkv'),
                call.send('<mediainfo for baz.mkv>'),
                call.emit('mediainfo', 'baz.mkv', '<mediainfo for baz.mkv>'),
            ],
            id='MKV / All / ContentError is reported',
        ),

        pytest.param(
            ('foo.mkv', 'bar.mkv', 'baz.mkv'),
            True,
            {
                'foo.mkv': None,
                'bar.mkv': Exception('NO'),
                'baz.mkv': None,
            },
            [
                call.mediainfo('foo.mkv'),
                call.send('<mediainfo for foo.mkv>'),
                call.emit('mediainfo', 'foo.mkv', '<mediainfo for foo.mkv>'),
                call.mediainfo('bar.mkv'),
                Exception('NO'),
            ],
            id='MKV / All / Exception is raised',
        ),

        pytest.param(
            ('foo/VIDEO_TS/VTS_01_2.VOB', 'bar/VIDEO_TS/VTS_03_4.VOB', 'baz/VIDEO_TS/VTS_05_6.VOB'),
            False,
            {
                '<IFO for foo/VIDEO_TS/VTS_01_2.VOB>': None,
                'foo/VIDEO_TS/VTS_01_2.VOB': None,
                '<IFO for bar/VIDEO_TS/VTS_03_4.VOB>': None,
                'bar/VIDEO_TS/VTS_03_4.VOB': None,
                '<IFO for baz/VIDEO_TS/VTS_05_6.VOB>': None,
                'baz/VIDEO_TS/VTS_05_6.VOB': None,
            },
            [
                call.vob2ifo('foo/VIDEO_TS/VTS_01_2.VOB'),
                call.mediainfo('<IFO for foo/VIDEO_TS/VTS_01_2.VOB>'),
                call.mediainfo('foo/VIDEO_TS/VTS_01_2.VOB'),
                call.send('<mediainfo for <IFO for foo/VIDEO_TS/VTS_01_2.VOB>>'),
                call.emit('mediainfo', '<IFO for foo/VIDEO_TS/VTS_01_2.VOB>', '<mediainfo for <IFO for foo/VIDEO_TS/VTS_01_2.VOB>>'),
                call.send('<mediainfo for foo/VIDEO_TS/VTS_01_2.VOB>'),
                call.emit('mediainfo', 'foo/VIDEO_TS/VTS_01_2.VOB', '<mediainfo for foo/VIDEO_TS/VTS_01_2.VOB>'),
            ],
            id='DVD / First / No exception',
        ),

        pytest.param(
            ('foo/VIDEO_TS/VTS_01_2.VOB', 'bar/VIDEO_TS/VTS_03_4.VOB', 'baz/VIDEO_TS/VTS_05_6.VOB'),
            False,
            {
                '<IFO for foo/VIDEO_TS/VTS_01_2.VOB>': None,
                'foo/VIDEO_TS/VTS_01_2.VOB': None,
                '<IFO for bar/VIDEO_TS/VTS_03_4.VOB>': errors.ContentError('no'),
                'bar/VIDEO_TS/VTS_03_4.VOB': None,
                '<IFO for baz/VIDEO_TS/VTS_05_6.VOB>': None,
                'baz/VIDEO_TS/VTS_05_6.VOB': None,
            },
            [
                call.vob2ifo('foo/VIDEO_TS/VTS_01_2.VOB'),
                call.mediainfo('<IFO for foo/VIDEO_TS/VTS_01_2.VOB>'),
                call.mediainfo('foo/VIDEO_TS/VTS_01_2.VOB'),
                call.send('<mediainfo for <IFO for foo/VIDEO_TS/VTS_01_2.VOB>>'),
                call.emit('mediainfo', '<IFO for foo/VIDEO_TS/VTS_01_2.VOB>', '<mediainfo for <IFO for foo/VIDEO_TS/VTS_01_2.VOB>>'),
                call.send('<mediainfo for foo/VIDEO_TS/VTS_01_2.VOB>'),
                call.emit('mediainfo', 'foo/VIDEO_TS/VTS_01_2.VOB', '<mediainfo for foo/VIDEO_TS/VTS_01_2.VOB>'),
            ],
            id='DVD / First / ContentError is irrelevant for IFO',
        ),

        pytest.param(
            ('foo/VIDEO_TS/VTS_01_2.VOB', 'bar/VIDEO_TS/VTS_03_4.VOB', 'baz/VIDEO_TS/VTS_05_6.VOB'),
            False,
            {
                '<IFO for foo/VIDEO_TS/VTS_01_2.VOB>': None,
                'foo/VIDEO_TS/VTS_01_2.VOB': None,
                '<IFO for bar/VIDEO_TS/VTS_03_4.VOB>': None,
                'bar/VIDEO_TS/VTS_03_4.VOB': errors.ContentError('no'),
                '<IFO for baz/VIDEO_TS/VTS_05_6.VOB>': None,
                'baz/VIDEO_TS/VTS_05_6.VOB': None,
            },
            [
                call.vob2ifo('foo/VIDEO_TS/VTS_01_2.VOB'),
                call.mediainfo('<IFO for foo/VIDEO_TS/VTS_01_2.VOB>'),
                call.mediainfo('foo/VIDEO_TS/VTS_01_2.VOB'),
                call.send('<mediainfo for <IFO for foo/VIDEO_TS/VTS_01_2.VOB>>'),
                call.emit('mediainfo', '<IFO for foo/VIDEO_TS/VTS_01_2.VOB>', '<mediainfo for <IFO for foo/VIDEO_TS/VTS_01_2.VOB>>'),
                call.send('<mediainfo for foo/VIDEO_TS/VTS_01_2.VOB>'),
                call.emit('mediainfo', 'foo/VIDEO_TS/VTS_01_2.VOB', '<mediainfo for foo/VIDEO_TS/VTS_01_2.VOB>'),
            ],
            id='DVD / First / ContentError is irrelevant for VOB',
        ),

        pytest.param(
            ('foo/VIDEO_TS/VTS_01_2.VOB', 'bar/VIDEO_TS/VTS_03_4.VOB', 'baz/VIDEO_TS/VTS_05_6.VOB'),
            False,
            {
                '<IFO for foo/VIDEO_TS/VTS_01_2.VOB>': errors.ContentError('no'),
                'foo/VIDEO_TS/VTS_01_2.VOB': None,
                '<IFO for bar/VIDEO_TS/VTS_03_4.VOB>': None,
                'bar/VIDEO_TS/VTS_03_4.VOB': None,
                '<IFO for baz/VIDEO_TS/VTS_05_6.VOB>': None,
                'baz/VIDEO_TS/VTS_05_6.VOB': None,
            },
            [
                call.vob2ifo('foo/VIDEO_TS/VTS_01_2.VOB'),
                call.mediainfo('<IFO for foo/VIDEO_TS/VTS_01_2.VOB>'),
                call.mediainfo('foo/VIDEO_TS/VTS_01_2.VOB'),
                call.error(errors.ContentError('no')),
                call.send('<mediainfo for foo/VIDEO_TS/VTS_01_2.VOB>'),
                call.emit('mediainfo', 'foo/VIDEO_TS/VTS_01_2.VOB', '<mediainfo for foo/VIDEO_TS/VTS_01_2.VOB>'),
            ],
            id='DVD / First / ContentError is reported for IFO',
        ),

        pytest.param(
            ('foo/VIDEO_TS/VTS_01_2.VOB', 'bar/VIDEO_TS/VTS_03_4.VOB', 'baz/VIDEO_TS/VTS_05_6.VOB'),
            False,
            {
                '<IFO for foo/VIDEO_TS/VTS_01_2.VOB>': None,
                'foo/VIDEO_TS/VTS_01_2.VOB': errors.ContentError('no'),
                '<IFO for bar/VIDEO_TS/VTS_03_4.VOB>': None,
                'bar/VIDEO_TS/VTS_03_4.VOB': None,
                '<IFO for baz/VIDEO_TS/VTS_05_6.VOB>': None,
                'baz/VIDEO_TS/VTS_05_6.VOB': None,
            },
            [
                call.vob2ifo('foo/VIDEO_TS/VTS_01_2.VOB'),
                call.mediainfo('<IFO for foo/VIDEO_TS/VTS_01_2.VOB>'),
                call.mediainfo('foo/VIDEO_TS/VTS_01_2.VOB'),
                call.send('<mediainfo for <IFO for foo/VIDEO_TS/VTS_01_2.VOB>>'),
                call.emit('mediainfo', '<IFO for foo/VIDEO_TS/VTS_01_2.VOB>', '<mediainfo for <IFO for foo/VIDEO_TS/VTS_01_2.VOB>>'),
                call.error(errors.ContentError('no')),
            ],
            id='DVD / First / ContentError is reported for VOB',
        ),

        pytest.param(
            ('foo/VIDEO_TS/VTS_01_2.VOB', 'bar/VIDEO_TS/VTS_03_4.VOB', 'baz/VIDEO_TS/VTS_05_6.VOB'),
            False,
            {
                '<IFO for foo/VIDEO_TS/VTS_01_2.VOB>': None,
                'foo/VIDEO_TS/VTS_01_2.VOB': None,
                '<IFO for bar/VIDEO_TS/VTS_03_4.VOB>': Exception('NO'),
                'bar/VIDEO_TS/VTS_03_4.VOB': None,
                '<IFO for baz/VIDEO_TS/VTS_05_6.VOB>': None,
                'baz/VIDEO_TS/VTS_05_6.VOB': None,
            },
            [
                call.vob2ifo('foo/VIDEO_TS/VTS_01_2.VOB'),
                call.mediainfo('<IFO for foo/VIDEO_TS/VTS_01_2.VOB>'),
                call.mediainfo('foo/VIDEO_TS/VTS_01_2.VOB'),
                call.send('<mediainfo for <IFO for foo/VIDEO_TS/VTS_01_2.VOB>>'),
                call.emit('mediainfo', '<IFO for foo/VIDEO_TS/VTS_01_2.VOB>', '<mediainfo for <IFO for foo/VIDEO_TS/VTS_01_2.VOB>>'),
                call.send('<mediainfo for foo/VIDEO_TS/VTS_01_2.VOB>'),
                call.emit('mediainfo', 'foo/VIDEO_TS/VTS_01_2.VOB', '<mediainfo for foo/VIDEO_TS/VTS_01_2.VOB>'),
            ],
            id='DVD / First / Exception is irrelevant for IFO',
        ),

        pytest.param(
            ('foo/VIDEO_TS/VTS_01_2.VOB', 'bar/VIDEO_TS/VTS_03_4.VOB', 'baz/VIDEO_TS/VTS_05_6.VOB'),
            False,
            {
                '<IFO for foo/VIDEO_TS/VTS_01_2.VOB>': None,
                'foo/VIDEO_TS/VTS_01_2.VOB': None,
                '<IFO for bar/VIDEO_TS/VTS_03_4.VOB>': None,
                'bar/VIDEO_TS/VTS_03_4.VOB': Exception('NO'),
                '<IFO for baz/VIDEO_TS/VTS_05_6.VOB>': None,
                'baz/VIDEO_TS/VTS_05_6.VOB': None,
            },
            [
                call.vob2ifo('foo/VIDEO_TS/VTS_01_2.VOB'),
                call.mediainfo('<IFO for foo/VIDEO_TS/VTS_01_2.VOB>'),
                call.mediainfo('foo/VIDEO_TS/VTS_01_2.VOB'),
                call.send('<mediainfo for <IFO for foo/VIDEO_TS/VTS_01_2.VOB>>'),
                call.emit('mediainfo', '<IFO for foo/VIDEO_TS/VTS_01_2.VOB>', '<mediainfo for <IFO for foo/VIDEO_TS/VTS_01_2.VOB>>'),
                call.send('<mediainfo for foo/VIDEO_TS/VTS_01_2.VOB>'),
                call.emit('mediainfo', 'foo/VIDEO_TS/VTS_01_2.VOB', '<mediainfo for foo/VIDEO_TS/VTS_01_2.VOB>'),
            ],
            id='DVD / First / Exception is irrelevant for VOB',
        ),

        pytest.param(
            ('foo/VIDEO_TS/VTS_01_2.VOB', 'bar/VIDEO_TS/VTS_03_4.VOB', 'baz/VIDEO_TS/VTS_05_6.VOB'),
            False,
            {
                '<IFO for foo/VIDEO_TS/VTS_01_2.VOB>': Exception('NO'),
                'foo/VIDEO_TS/VTS_01_2.VOB': None,
                '<IFO for bar/VIDEO_TS/VTS_03_4.VOB>': None,
                'bar/VIDEO_TS/VTS_03_4.VOB': None,
                '<IFO for baz/VIDEO_TS/VTS_05_6.VOB>': None,
                'baz/VIDEO_TS/VTS_05_6.VOB': None,
            },
            [
                call.vob2ifo('foo/VIDEO_TS/VTS_01_2.VOB'),
                call.mediainfo('<IFO for foo/VIDEO_TS/VTS_01_2.VOB>'),
                call.mediainfo('foo/VIDEO_TS/VTS_01_2.VOB'),
                Exception('NO'),
            ],
            id='DVD / First / Exception is raised for IFO',
        ),

        pytest.param(
            ('foo/VIDEO_TS/VTS_01_2.VOB', 'bar/VIDEO_TS/VTS_03_4.VOB', 'baz/VIDEO_TS/VTS_05_6.VOB'),
            False,
            {
                '<IFO for foo/VIDEO_TS/VTS_01_2.VOB>': None,
                'foo/VIDEO_TS/VTS_01_2.VOB': Exception('NO'),
                '<IFO for bar/VIDEO_TS/VTS_03_4.VOB>': None,
                'bar/VIDEO_TS/VTS_03_4.VOB': None,
                '<IFO for baz/VIDEO_TS/VTS_05_6.VOB>': None,
                'baz/VIDEO_TS/VTS_05_6.VOB': None,
            },
            [
                call.vob2ifo('foo/VIDEO_TS/VTS_01_2.VOB'),
                call.mediainfo('<IFO for foo/VIDEO_TS/VTS_01_2.VOB>'),
                call.mediainfo('foo/VIDEO_TS/VTS_01_2.VOB'),
                call.send('<mediainfo for <IFO for foo/VIDEO_TS/VTS_01_2.VOB>>'),
                call.emit('mediainfo', '<IFO for foo/VIDEO_TS/VTS_01_2.VOB>', '<mediainfo for <IFO for foo/VIDEO_TS/VTS_01_2.VOB>>'),
                Exception('NO'),
            ],
            id='DVD / First / Exception is raised for VOB',
        ),

        pytest.param(
            ('foo/VIDEO_TS/VTS_01_2.VOB', 'bar/VIDEO_TS/VTS_03_4.VOB', 'baz/VIDEO_TS/VTS_05_6.VOB'),
            True,
            {
                '<IFO for foo/VIDEO_TS/VTS_01_2.VOB>': None,
                'foo/VIDEO_TS/VTS_01_2.VOB': None,
                '<IFO for bar/VIDEO_TS/VTS_03_4.VOB>': None,
                'bar/VIDEO_TS/VTS_03_4.VOB': None,
                '<IFO for baz/VIDEO_TS/VTS_05_6.VOB>': None,
                'baz/VIDEO_TS/VTS_05_6.VOB': None,
            },
            [
                call.vob2ifo('foo/VIDEO_TS/VTS_01_2.VOB'),
                call.mediainfo('<IFO for foo/VIDEO_TS/VTS_01_2.VOB>'),
                call.mediainfo('foo/VIDEO_TS/VTS_01_2.VOB'),
                call.send('<mediainfo for <IFO for foo/VIDEO_TS/VTS_01_2.VOB>>'),
                call.emit('mediainfo', '<IFO for foo/VIDEO_TS/VTS_01_2.VOB>', '<mediainfo for <IFO for foo/VIDEO_TS/VTS_01_2.VOB>>'),
                call.send('<mediainfo for foo/VIDEO_TS/VTS_01_2.VOB>'),
                call.emit('mediainfo', 'foo/VIDEO_TS/VTS_01_2.VOB', '<mediainfo for foo/VIDEO_TS/VTS_01_2.VOB>'),

                call.vob2ifo('bar/VIDEO_TS/VTS_03_4.VOB'),
                call.mediainfo('<IFO for bar/VIDEO_TS/VTS_03_4.VOB>'),
                call.mediainfo('bar/VIDEO_TS/VTS_03_4.VOB'),
                call.send('<mediainfo for <IFO for bar/VIDEO_TS/VTS_03_4.VOB>>'),
                call.emit('mediainfo', '<IFO for bar/VIDEO_TS/VTS_03_4.VOB>', '<mediainfo for <IFO for bar/VIDEO_TS/VTS_03_4.VOB>>'),
                call.send('<mediainfo for bar/VIDEO_TS/VTS_03_4.VOB>'),
                call.emit('mediainfo', 'bar/VIDEO_TS/VTS_03_4.VOB', '<mediainfo for bar/VIDEO_TS/VTS_03_4.VOB>'),

                call.vob2ifo('baz/VIDEO_TS/VTS_05_6.VOB'),
                call.mediainfo('<IFO for baz/VIDEO_TS/VTS_05_6.VOB>'),
                call.mediainfo('baz/VIDEO_TS/VTS_05_6.VOB'),
                call.send('<mediainfo for <IFO for baz/VIDEO_TS/VTS_05_6.VOB>>'),
                call.emit('mediainfo', '<IFO for baz/VIDEO_TS/VTS_05_6.VOB>', '<mediainfo for <IFO for baz/VIDEO_TS/VTS_05_6.VOB>>'),
                call.send('<mediainfo for baz/VIDEO_TS/VTS_05_6.VOB>'),
                call.emit('mediainfo', 'baz/VIDEO_TS/VTS_05_6.VOB', '<mediainfo for baz/VIDEO_TS/VTS_05_6.VOB>'),
            ],
            id='DVD / All / No exception',
        ),

        pytest.param(
            ('foo/VIDEO_TS/VTS_01_2.VOB', 'bar/VIDEO_TS/VTS_03_4.VOB', 'baz/VIDEO_TS/VTS_05_6.VOB'),
            True,
            {
                '<IFO for foo/VIDEO_TS/VTS_01_2.VOB>': None,
                'foo/VIDEO_TS/VTS_01_2.VOB': None,
                '<IFO for bar/VIDEO_TS/VTS_03_4.VOB>': errors.ContentError('no'),
                'bar/VIDEO_TS/VTS_03_4.VOB': None,
                '<IFO for baz/VIDEO_TS/VTS_05_6.VOB>': None,
                'baz/VIDEO_TS/VTS_05_6.VOB': None,
            },
            [
                call.vob2ifo('foo/VIDEO_TS/VTS_01_2.VOB'),
                call.mediainfo('<IFO for foo/VIDEO_TS/VTS_01_2.VOB>'),
                call.mediainfo('foo/VIDEO_TS/VTS_01_2.VOB'),
                call.send('<mediainfo for <IFO for foo/VIDEO_TS/VTS_01_2.VOB>>'),
                call.emit('mediainfo', '<IFO for foo/VIDEO_TS/VTS_01_2.VOB>', '<mediainfo for <IFO for foo/VIDEO_TS/VTS_01_2.VOB>>'),
                call.send('<mediainfo for foo/VIDEO_TS/VTS_01_2.VOB>'),
                call.emit('mediainfo', 'foo/VIDEO_TS/VTS_01_2.VOB', '<mediainfo for foo/VIDEO_TS/VTS_01_2.VOB>'),

                call.vob2ifo('bar/VIDEO_TS/VTS_03_4.VOB'),
                call.mediainfo('<IFO for bar/VIDEO_TS/VTS_03_4.VOB>'),
                call.mediainfo('bar/VIDEO_TS/VTS_03_4.VOB'),
                call.error(errors.ContentError('no')),
                call.send('<mediainfo for bar/VIDEO_TS/VTS_03_4.VOB>'),
                call.emit('mediainfo', 'bar/VIDEO_TS/VTS_03_4.VOB', '<mediainfo for bar/VIDEO_TS/VTS_03_4.VOB>'),

                call.vob2ifo('baz/VIDEO_TS/VTS_05_6.VOB'),
                call.mediainfo('<IFO for baz/VIDEO_TS/VTS_05_6.VOB>'),
                call.mediainfo('baz/VIDEO_TS/VTS_05_6.VOB'),
                call.send('<mediainfo for <IFO for baz/VIDEO_TS/VTS_05_6.VOB>>'),
                call.emit('mediainfo', '<IFO for baz/VIDEO_TS/VTS_05_6.VOB>', '<mediainfo for <IFO for baz/VIDEO_TS/VTS_05_6.VOB>>'),
                call.send('<mediainfo for baz/VIDEO_TS/VTS_05_6.VOB>'),
                call.emit('mediainfo', 'baz/VIDEO_TS/VTS_05_6.VOB', '<mediainfo for baz/VIDEO_TS/VTS_05_6.VOB>'),
            ],
            id='DVD / All / ContentError is reported for IFO',
        ),

        pytest.param(
            ('foo/VIDEO_TS/VTS_01_2.VOB', 'bar/VIDEO_TS/VTS_03_4.VOB', 'baz/VIDEO_TS/VTS_05_6.VOB'),
            True,
            {
                '<IFO for foo/VIDEO_TS/VTS_01_2.VOB>': None,
                'foo/VIDEO_TS/VTS_01_2.VOB': None,
                '<IFO for bar/VIDEO_TS/VTS_03_4.VOB>': None,
                'bar/VIDEO_TS/VTS_03_4.VOB': errors.ContentError('no'),
                '<IFO for baz/VIDEO_TS/VTS_05_6.VOB>': None,
                'baz/VIDEO_TS/VTS_05_6.VOB': None,
            },
            [
                call.vob2ifo('foo/VIDEO_TS/VTS_01_2.VOB'),
                call.mediainfo('<IFO for foo/VIDEO_TS/VTS_01_2.VOB>'),
                call.mediainfo('foo/VIDEO_TS/VTS_01_2.VOB'),
                call.send('<mediainfo for <IFO for foo/VIDEO_TS/VTS_01_2.VOB>>'),
                call.emit('mediainfo', '<IFO for foo/VIDEO_TS/VTS_01_2.VOB>', '<mediainfo for <IFO for foo/VIDEO_TS/VTS_01_2.VOB>>'),
                call.send('<mediainfo for foo/VIDEO_TS/VTS_01_2.VOB>'),
                call.emit('mediainfo', 'foo/VIDEO_TS/VTS_01_2.VOB', '<mediainfo for foo/VIDEO_TS/VTS_01_2.VOB>'),

                call.vob2ifo('bar/VIDEO_TS/VTS_03_4.VOB'),
                call.mediainfo('<IFO for bar/VIDEO_TS/VTS_03_4.VOB>'),
                call.mediainfo('bar/VIDEO_TS/VTS_03_4.VOB'),
                call.send('<mediainfo for <IFO for bar/VIDEO_TS/VTS_03_4.VOB>>'),
                call.emit('mediainfo', '<IFO for bar/VIDEO_TS/VTS_03_4.VOB>', '<mediainfo for <IFO for bar/VIDEO_TS/VTS_03_4.VOB>>'),
                call.error(errors.ContentError('no')),

                call.vob2ifo('baz/VIDEO_TS/VTS_05_6.VOB'),
                call.mediainfo('<IFO for baz/VIDEO_TS/VTS_05_6.VOB>'),
                call.mediainfo('baz/VIDEO_TS/VTS_05_6.VOB'),
                call.send('<mediainfo for <IFO for baz/VIDEO_TS/VTS_05_6.VOB>>'),
                call.emit('mediainfo', '<IFO for baz/VIDEO_TS/VTS_05_6.VOB>', '<mediainfo for <IFO for baz/VIDEO_TS/VTS_05_6.VOB>>'),
                call.send('<mediainfo for baz/VIDEO_TS/VTS_05_6.VOB>'),
                call.emit('mediainfo', 'baz/VIDEO_TS/VTS_05_6.VOB', '<mediainfo for baz/VIDEO_TS/VTS_05_6.VOB>'),
            ],
            id='DVD / All / ContentError is reported for VOB',
        ),

        pytest.param(
            ('foo/VIDEO_TS/VTS_01_2.VOB', 'bar/VIDEO_TS/VTS_03_4.VOB', 'baz/VIDEO_TS/VTS_05_6.VOB'),
            True,
            {
                '<IFO for foo/VIDEO_TS/VTS_01_2.VOB>': None,
                'foo/VIDEO_TS/VTS_01_2.VOB': None,
                '<IFO for bar/VIDEO_TS/VTS_03_4.VOB>': Exception('NO'),
                'bar/VIDEO_TS/VTS_03_4.VOB': None,
                '<IFO for baz/VIDEO_TS/VTS_05_6.VOB>': None,
                'baz/VIDEO_TS/VTS_05_6.VOB': None,
            },
            [
                call.vob2ifo('foo/VIDEO_TS/VTS_01_2.VOB'),
                call.mediainfo('<IFO for foo/VIDEO_TS/VTS_01_2.VOB>'),
                call.mediainfo('foo/VIDEO_TS/VTS_01_2.VOB'),
                call.send('<mediainfo for <IFO for foo/VIDEO_TS/VTS_01_2.VOB>>'),
                call.emit('mediainfo', '<IFO for foo/VIDEO_TS/VTS_01_2.VOB>', '<mediainfo for <IFO for foo/VIDEO_TS/VTS_01_2.VOB>>'),
                call.send('<mediainfo for foo/VIDEO_TS/VTS_01_2.VOB>'),
                call.emit('mediainfo', 'foo/VIDEO_TS/VTS_01_2.VOB', '<mediainfo for foo/VIDEO_TS/VTS_01_2.VOB>'),

                call.vob2ifo('bar/VIDEO_TS/VTS_03_4.VOB'),
                call.mediainfo('<IFO for bar/VIDEO_TS/VTS_03_4.VOB>'),
                call.mediainfo('bar/VIDEO_TS/VTS_03_4.VOB'),
                Exception('NO'),
            ],
            id='DVD / All / Exception is raised for IFO',
        ),

        pytest.param(
            ('foo/VIDEO_TS/VTS_01_2.VOB', 'bar/VIDEO_TS/VTS_03_4.VOB', 'baz/VIDEO_TS/VTS_05_6.VOB'),
            True,
            {
                '<IFO for foo/VIDEO_TS/VTS_01_2.VOB>': None,
                'foo/VIDEO_TS/VTS_01_2.VOB': None,
                '<IFO for bar/VIDEO_TS/VTS_03_4.VOB>': None,
                'bar/VIDEO_TS/VTS_03_4.VOB': Exception('NO'),
                '<IFO for baz/VIDEO_TS/VTS_05_6.VOB>': None,
                'baz/VIDEO_TS/VTS_05_6.VOB': None,
            },
            [
                call.vob2ifo('foo/VIDEO_TS/VTS_01_2.VOB'),
                call.mediainfo('<IFO for foo/VIDEO_TS/VTS_01_2.VOB>'),
                call.mediainfo('foo/VIDEO_TS/VTS_01_2.VOB'),
                call.send('<mediainfo for <IFO for foo/VIDEO_TS/VTS_01_2.VOB>>'),
                call.emit('mediainfo', '<IFO for foo/VIDEO_TS/VTS_01_2.VOB>', '<mediainfo for <IFO for foo/VIDEO_TS/VTS_01_2.VOB>>'),
                call.send('<mediainfo for foo/VIDEO_TS/VTS_01_2.VOB>'),
                call.emit('mediainfo', 'foo/VIDEO_TS/VTS_01_2.VOB', '<mediainfo for foo/VIDEO_TS/VTS_01_2.VOB>'),

                call.vob2ifo('bar/VIDEO_TS/VTS_03_4.VOB'),
                call.mediainfo('<IFO for bar/VIDEO_TS/VTS_03_4.VOB>'),
                call.mediainfo('bar/VIDEO_TS/VTS_03_4.VOB'),
                call.send('<mediainfo for <IFO for bar/VIDEO_TS/VTS_03_4.VOB>>'),
                call.emit('mediainfo', '<IFO for bar/VIDEO_TS/VTS_03_4.VOB>', '<mediainfo for <IFO for bar/VIDEO_TS/VTS_03_4.VOB>>'),
                Exception('NO'),
            ],
            id='DVD / All / Exception is raised for VOB',
        ),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.asyncio
async def test_generate_mediainfos(filtered_videos, from_all_videos, mediainfo_exceptions, exp_mock_calls,
                                   job, mocker):
    job._from_all_videos = from_all_videos
    mocks = Mock()
    mocks.attach_mock(
        mocker.patch('upsies.utils.torrent.filter_files'),
        'filter_files',
    )
    mocks.attach_mock(
        mocker.patch('upsies.utils.video.filter_main_videos', return_value=filtered_videos),
        'filter_main_videos',
    )

    def mediainfo_side_effect(filepath):
        exception = mediainfo_exceptions.get(filepath, None)
        print('side_effect for', filepath, 'is', repr(exception))
        if exception:
            raise exception
        else:
            return f'<mediainfo for {filepath}>'

    mocks.attach_mock(
        mocker.patch('upsies.utils.video.mediainfo', side_effect=mediainfo_side_effect),
        'mediainfo',
    )
    mocks.attach_mock(
        mocker.patch('upsies.utils.video.vob2ifo', side_effect=lambda f: f'<IFO for {f}>'),
        'vob2ifo',
    )
    mocks.attach_mock(mocker.patch.object(job, 'send'), 'send')
    mocks.attach_mock(mocker.patch.object(job, 'error'), 'error')
    mocks.attach_mock(mocker.patch.object(job.signal, 'emit'), 'emit')

    exp_exceptions = [x for x in exp_mock_calls if isinstance(x, Exception)]
    exp_mock_calls = [x for x in exp_mock_calls if not isinstance(x, Exception)]

    if exp_exceptions:
        exp_exception = exp_exceptions[0]
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            await job._generate_mediainfos()
    else:
        await job._generate_mediainfos()

    assert mocks.mock_calls == [
        call.filter_files(
            content_path=job._content_path,
            exclude=job._exclude_files,
        ),
        call.filter_main_videos(mocks.filter_files.return_value),

    ] + exp_mock_calls


def test_MediainfoJob_store_mediainfos_by_file(job):
    assert job._mediainfos_by_file == {}

    job._store_mediainfos_by_file('foo.mkv', 'mediainfo for foo.mkv')
    assert job._mediainfos_by_file == {
        'foo.mkv': 'mediainfo for foo.mkv',
    }

    job._store_mediainfos_by_file('bar.mkv', 'mediainfo for bar.mkv')
    assert job._mediainfos_by_file == {
        'foo.mkv': 'mediainfo for foo.mkv',
        'bar.mkv': 'mediainfo for bar.mkv',
    }

    job._store_mediainfos_by_file('dvd/VIDEO_TS/VTS_01_0.IFO', 'mediainfo for VTS_01_0.IFO')
    assert job._mediainfos_by_file == {
        'foo.mkv': 'mediainfo for foo.mkv',
        'bar.mkv': 'mediainfo for bar.mkv',
        'dvd/VIDEO_TS/VTS_01_0.IFO': 'mediainfo for VTS_01_0.IFO',
    }
    job._store_mediainfos_by_file('dvd/VIDEO_TS/VTS_01_1.VOB', 'mediainfo for VTS_01_1.VOB')
    assert job._mediainfos_by_file == {
        'foo.mkv': 'mediainfo for foo.mkv',
        'bar.mkv': 'mediainfo for bar.mkv',
        'dvd/VIDEO_TS/VTS_01_0.IFO': 'mediainfo for VTS_01_0.IFO',
        'dvd/VIDEO_TS/VTS_01_1.VOB': 'mediainfo for VTS_01_1.VOB',
    }


def test_MediainfoJob_mediainfos_by_file(job):
    job._mediainfos_by_file = {
        'path/to/foo.mkv': 'mediainfo for foo.mkv',
    }

    mediainfos_by_file = job.mediainfos_by_file
    assert mediainfos_by_file == {
        'path/to/foo.mkv': 'mediainfo for foo.mkv',
    }

    mediainfos_by_file['path/to/foo.mkv'] = 'not a mediainfo'
    assert job.mediainfos_by_file == {
        'path/to/foo.mkv': 'mediainfo for foo.mkv',
    }

    mediainfos_by_file['path/to/bar.mkv'] = 'mediainfo for bar.mkv'
    assert job.mediainfos_by_file == {
        'path/to/foo.mkv': 'mediainfo for foo.mkv',
    }

    del mediainfos_by_file['path/to/foo.mkv']
    assert job.mediainfos_by_file == {
        'path/to/foo.mkv': 'mediainfo for foo.mkv',
    }



import re
from unittest.mock import AsyncMock, Mock, call

import pytest

from upsies.utils.webdbs.base import WebDbApiBase
from upsies.utils.webdbs.common import Query


def make_TestWebDbApi(default_config=None, **kwargs):
    # Avoid NameError bug (https://github.com/python/cpython/issues/87546)
    default_config_ = default_config

    class TestWebDbApi(WebDbApiBase):
        name = 'foo'
        label = 'Foo'
        default_config = default_config_ or {}
        search = AsyncMock()
        get_id_from_text = Mock(return_value=None)

        directors = AsyncMock()
        creators = AsyncMock()
        cast = AsyncMock()
        _countries = AsyncMock()
        genres = AsyncMock()
        poster_url = AsyncMock()
        rating_min = 0.0
        rating_max = 10.0
        rating = AsyncMock()
        _runtimes = AsyncMock()
        summary = AsyncMock()
        _titles_english = AsyncMock()
        _title_original = AsyncMock()
        type = AsyncMock()
        url = AsyncMock()
        year = AsyncMock()

    return TestWebDbApi(**kwargs)

@pytest.fixture
def webdb():
    return make_TestWebDbApi()


def test_name_property(webdb):
    assert webdb.name == 'foo'


def test_label_property(webdb):
    assert webdb.label == 'Foo'


def test_no_results_info_property(webdb):
    assert webdb.no_results_info == ''


def test_config_property():
    webdb = make_TestWebDbApi()
    assert webdb.config == {}
    webdb = make_TestWebDbApi(default_config={'foo': 1, 'bar': 2})
    assert webdb.config == {'foo': 1, 'bar': 2}
    webdb = make_TestWebDbApi(default_config={'foo': 1, 'bar': 2}, config={'bar': 99})
    assert webdb.config == {'foo': 1, 'bar': 99}


@pytest.mark.parametrize(
    argnames='query, found_id, exp_result',
    argvalues=(
        pytest.param(
            123,
            None,
            TypeError('Not a Query instance: 123'),
            id='Bad query object',
        ),
        pytest.param(
            Query('The Foo', type='movie', year='2000'),
            None,
            Query('The Foo', type='movie', year='2000'),
            id='No ID found in query title',
        ),
        pytest.param(
            Query('The Foo', type='movie', year='2000'),
            'abc.123',
            Query('The Foo', type='movie', year='2000', id='abc.123'),
            id='ID is found in query title',
        ),
    ),
    ids=lambda v: repr(v),
)
def test_sanitize_query(query, found_id, exp_result, webdb, mocker):
    mocker.patch.object(webdb, 'get_id_from_text', return_value=found_id)
    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            webdb.sanitize_query(query)
    else:
        return_value = webdb.sanitize_query(query)
        assert return_value == exp_result
        assert webdb.get_id_from_text.call_args_list == [call(query.title)]


@pytest.mark.parametrize(
    argnames='countries, exp_countries',
    argvalues=(
        (('usa', 'Peoples Republic of China'), ('United States', 'China')),
    ),
)
@pytest.mark.asyncio
async def test_countries(countries, exp_countries, webdb, mocker):
    mocker.patch.object(webdb, '_countries', AsyncMock(return_value=countries))
    return_value = await webdb.countries('mock id')
    assert return_value == exp_countries
    assert webdb._countries.call_args_list == [call('mock id')]


@pytest.mark.parametrize(
    argnames='runtimes, exp_runtimes',
    argvalues=(
        ({'usa': '123', 'Peoples Republic of China': '456', 'default': '789'},
         {'United States': '123', 'China': '456', 'default': '789'}),
        ({'usa': '123', 'Peoples Republic of China': '456', 'default': '123'},
         {'United States': '123', 'China': '456'}),
    ),
)
@pytest.mark.asyncio
async def test_runtimes(runtimes, exp_runtimes, webdb, mocker):
    mocker.patch.object(webdb, '_runtimes', AsyncMock(return_value=runtimes))
    return_value = await webdb.runtimes('mock id')
    assert return_value == exp_runtimes
    assert webdb._runtimes.call_args_list == [call('mock id')]


@pytest.mark.asyncio
async def test_gather(webdb):
    webdb.cast.return_value = 'mock cast'
    webdb.genres.return_value = ['mock genre one', 'mock genre two']
    webdb.summary.return_value = 'mock summary'
    webdb.type.return_value = 'mock type'
    webdb.year.return_value = 'mock year'
    assert await webdb.gather('mock id', 'type', 'year', 'cast', 'genres') == {
        'cast': 'mock cast',
        'id': 'mock id',
        'genres': ['mock genre one', 'mock genre two'],
        'type': 'mock type',
        'year': 'mock year',
    }
    assert webdb.cast.call_args_list == [call('mock id')]
    assert webdb._countries.call_args_list == []
    assert webdb.summary.call_args_list == []
    assert webdb.type.call_args_list == [call('mock id')]
    assert webdb.year.call_args_list == [call('mock id')]



"""
Subtitles information
"""

import re

from .. import utils
from . import LazyModule

import logging  # isort:skip
_log = logging.getLogger(__name__)

langcodes = LazyModule(module='langcodes', namespace=globals())


# Map various strings to canonical sutitle format names
_subtitle_formats = {
    # Matroska Codec ID
    # Reference: https://www.matroska.org/technical/codec_specs.html
    'S_HDMV/PGS': 'PGS',     # Blu-ray
    'S_VOBSUB': 'VobSub',    # DVD
    'S_TEXT/UTF8': 'SRT',    # SubRip
    'S_TEXT/SSA': 'SSA',     # SubStation Alpha
    'S_TEXT/ASS': 'ASS',     # Advanced SubStation Alpha
    'S_TEXT/WEBVTT': 'VTT',  # Web Video Text Tracks

    # File extension
    'srt': 'SRT',     # SubRip
    'ssa': 'SSA',     # SubStation Alpha
    'ass': 'ASS',     # Advanced SubStation Alpha
    'vtt': 'VTT',     # Web Video Text Tracks
    'idx': 'VobSub',  # DVD
    'sub': 'VobSub',  # DVD
}


class Subtitle:
    """
    Dataclass that stores information about a subtitle track

    :param language: BCP47 language code with optional country code (e.g. "en",
        "en-UK", "es-419", "zh-HANS")

        References:
            https://www.matroska.org/technical/notes.html [Language Codes]
            https://en.wikipedia.org/wiki/IETF_language_tag

    :param bool forced: Whether the subtitle track should always be played
    :param bool format: Subtitle format (e.g. "SRT", "VobSub", "PGS")

    :raise ValueError: if `language` is invalid
    """

    def __init__(self, *, language, forced, format):
        # Parse and normalize `language`
        try:
            lc = langcodes.Language.get(language)
        except ValueError:
            self.language = ''
            self.territory = ''
        else:
            self.language = lc.language
            self.territory = lc.territory or ''

        self.forced = bool(forced)
        self.format = str(format or '')

    def __repr__(self):
        if self.territory:
            language = f'{self.language}-{self.territory}'
        else:
            language = self.language
        return f'{type(self).__name__}(language={language!r}, forced={self.forced!r}, format={self.format!r})'

    def __hash__(self):
        return hash((self.language, self.territory, self.forced, self.format))

    def __eq__(self, other):
        return hash(self) == hash(other)


def get_subtitles(content_path):
    """
    Return sequence of :class:`Subtitle` objects from all subtitle sources

    This function combines all ``get_subtitles_*`` functions.

    :param content_path: Path to release file or directory
    """
    subtitles = set()
    subtitles.update(get_subtitles_from_mediainfo(content_path))
    subtitles.update(get_subtitles_from_text_files(content_path))
    subtitles.update(get_subtitles_from_idx_files(content_path))
    subtitles.update(get_subtitles_from_dvd_tree(content_path))
    return subtitles


def get_subtitles_from_mediainfo(content_path):
    """
    Return sequence of :class:`Subtitle` objects from ``mediainfo`` output

    Only the first video is used to find subtitles (see :func:`find_videos`).

    :param content_path: Path to release file or directory
    """
    subtitles = set()
    all_tracks = utils.video.tracks(content_path)
    try:
        subtitle_tracks = all_tracks['Text']
    except KeyError:
        pass
    else:
        for track in subtitle_tracks:
            subtitle = Subtitle(
                language=track.get('Language', ''),
                forced=(track.get('Forced') == 'Yes'),
                format=_subtitle_formats.get(track.get('CodecID')),
            )
            _log.debug('Found subtitle language in mediainfo: %r', subtitle)
            subtitles.add(subtitle)
    _log.debug('Subtitles from mediainfo: %r', subtitles)
    return subtitles


_srt_filename_regex = re.compile(r'(?:\.([a-z]{2,3})|)\.(\w+)$')

def get_subtitles_from_text_files(content_path):
    """
    Return sequence of :class:`Subtitle` objects from subtitle file names

    For a video file named "foo.mkv", the subtitle language can be in the file
    name, e.g. "foo.en.srt" or "foo.eng.ass".

    If `content_path` is a directory, it is searched recursively for subtitle
    files.

    If `content_path` is not a directory, the returned sequence is empty.

    :param content_path: Path to release file or directory
    """
    subtitles = set()
    for filepath in utils.fs.file_list(content_path, extensions=('srt', 'ssa', 'ass', 'vtt')):
        filename = utils.fs.basename(filepath)
        match = _srt_filename_regex.search(filename)

        if match:
            language = match.group(1) or ''
            extension = match.group(2).lower()
            format = _subtitle_formats.get(extension, '')
        else:
            language = ''
            format = ''

        subtitle = Subtitle(
            language=language,
            forced=False,
            format=format,
        )
        _log.debug('Found text subtitle: %s: %r', filepath, subtitle)
        subtitles.add(subtitle)

    _log.debug('Subtitles from text files: %r', subtitles)
    return subtitles


def get_subtitles_from_idx_files(content_path):
    """
    Return sequence of :class:`Subtitle` objects from ``*.idx`` files

    For .idx/.sub pairs, the .idx file can contain the language.

    If `content_path` is a directory, it is searched recursively for ``*.idx``
    files.

    If `content_path` is not a directory, the returned sequence is empty.

    :param content_path: Path to release file or directory
    """
    subtitles = set()
    for idx_filepath in utils.fs.file_list(content_path, extensions=('idx',)):
        try:
            language = _get_language_code_from_idx_file(idx_filepath)
        except OSError as e:
            _log.debug('Ignoring exception from reading %r: %r', idx_filepath, e)
        else:
            subtitle = Subtitle(
                language=language,
                forced=False,
                format=_subtitle_formats['idx'],
            )
            _log.debug('Found IDX subtitle: %s: %r', idx_filepath, subtitle)
            subtitles.add(subtitle)
    _log.debug('Subtitles from idx files: %r', subtitles)
    return subtitles


_idx_language_regex = re.compile(r'^id:\s*([a-z]{2,3})\b')

def _get_language_code_from_idx_file(idx_filepath):
    # Example: "id: de, index: 0"
    # Expect 2-letter and 3-letter codes.
    with open(idx_filepath, 'r') as f:
        for line in f.readlines():
            match = _idx_language_regex.search(line)
            if match:
                language = match.group(1)
                _log.debug('Found subtitle language in %s: %r', idx_filepath, language)
                return language
    return ''


def get_subtitles_from_dvd_tree(content_path):
    """
    Return sequence of :class:`Subtitle` objects from ``VIDEO_TS``
    subdirectory

    If `content_path` is not a directory or does not contain a ``VIDEO_TS``
    subdirectory, the returned sequence is empty.

    :param content_path: Path directory that contains ``VIDEO_TS`` subdirectory
    """
    subtitles = set()
    main_videos = utils.video.find_videos(content_path)
    for main_video in main_videos:
        if utils.fs.file_extension(main_video) == 'vob':
            ifo_filepath = utils.video.vob2ifo(main_video)
            for sub in get_subtitles_from_mediainfo(ifo_filepath):
                sub.format = _subtitle_formats['S_VOBSUB']
                subtitles.add(sub)
    _log.debug('Subtitles from DVD tree(s): %r', subtitles)
    return subtitles



"""
Abstract base class for jobs
"""

import abc
import asyncio
import collections
import functools
import os
import pickle
import re

import unidecode

from .. import constants, utils

import logging  # isort:skip
_log = logging.getLogger(__name__)


# TODO: Split each job into tasks.
#
# 1. Rename attach_task() to add_task().
#
# 2. Every job must add at least one task in initialize().
#
# 3. A job finishes when all tasks are done.
#
# 4. A job can add or cancel tasks at any time, except when it is finished.
#
# 5. finish() can now be removed.
#
# 6. Implement utils.subproc.run_async() and replace utils.subproc.run() with it
#    where possible.
#
# 7. QueueJob is now no longer needed. Other jobs can add tasks to another job,
#    e.g. for uploading generated screenshots while screenshots are being
#    uploaded.
#
#    IMPORTANT: The job that uploads screenshots must know beforehand how many
#               screenshots to expect so it doesn't finish prematurely if
#               screenshot uploding is faster than generation.
#
# 8. Maybe remove utils.run_task().


class JobBase(abc.ABC):
    """
    Base class for all jobs

    :param str home_directory: Directory that is used to store created files
    :param str cache_directory: Directory that is used to cache output
    :param str ignore_cache: Whether cached output and previously created files
        should not be re-used
    :param bool no_output_is_ok: Whether the job can succeed without any
        :attr:`output`
    :param bool hidden: Whether to hide the job's output in the UI
    :param bool autostart: Whether this job is started automatically
    :param precondition: Callable that gets no arguments and returns whether
        this job is enabled or disabled
    :param prejobs: Sequence of :attr:`prerequisite jobs <prejobs>`
    :param presignals: Mapping of :class:`jobs <JobBase>` to sequence of signals
        (see :meth:`presignal`)
    :param callbacks: Mapping of :attr:`signal` names to callable or sequence of
        callables to :meth:`~.signal.Signal.register` for that signal

    Any additional keyword arguments are passed on to :meth:`initialize`.

    If possible, arguments should be validated before creating a job to fail
    early when sibling jobs haven't started doing work that has to be cancelled
    in case of error.

    Methods and properties of jobs should not raise any expected exceptions.
    Exceptions should instead be passed to :meth:`error` (report message to
    user) or to :meth:`exception` (throw traceback at user), and the job should
    :meth:`finish` immediately with an :attr:`exit_code` `> 0`.
    """

    @property
    @abc.abstractmethod
    def name(self):
        """Internal name (e.g. for the cache file name)"""

    @property
    @abc.abstractmethod
    def label(self):
        """User-facing name"""

    @property
    def home_directory(self):
        """
        Directory that is used to store files (e.g. generated images) or empty
        string

        This directory is guaranteed to exist.
        """
        if self._home_directory and not os.path.exists(self._home_directory):
            utils.fs.mkdir(self._home_directory)
        return self._home_directory

    @functools.cached_property
    def cache_directory(self):
        """
        Path to existing directory that stores :attr:`cache_file`

        This directory is guaranteed to exist.
        """
        if not os.path.exists(self._cache_directory):
            utils.fs.mkdir(self._cache_directory)
        return self._cache_directory

    @property
    def ignore_cache(self):
        """Whether cached output and previously created files should not be re-used"""
        return self._ignore_cache

    @property
    def no_output_is_ok(self):
        """Whether the job can succeed without any :attr:`output`"""
        return self._no_output_is_ok

    @property
    def hidden(self):
        """Whether to hide this job's output in the UI"""
        return self._hidden

    @property
    def autostart(self):
        """
        Whether this job is started automatically

        If this value is falsy, :meth:`start` must be called manually.

        See also :attr:`is_enabled`.
        """
        return self._autostart

    @property
    def precondition(self):
        """
        Callable that gets no arguments and returns whether this job should
        be :meth:`started <start>`

        See also :attr:`is_enabled`.
        """
        return self._precondition

    @precondition.setter
    def precondition(self, function):
        if self.is_started:
            raise RuntimeError('Cannot set precondition after job has been started')

        if not callable(function):
            raise TypeError(f'Not callable: {function!r}')
        self._precondition = function

    @property
    def prejobs(self):
        """
        Sequence of prerequisite :class:`jobs <upsies.jobs.base.JobBase>`

        All prejobs must be either :attr:`finished <is_finished>` or
        :attr:`disabled <is_enabled>` before this job can start.

        See also :attr:`is_enabled`.
        """
        return self._prejobs

    @prejobs.setter
    def prejobs(self, prejobs):
        if self.is_started:
            raise RuntimeError('Cannot set prejobs after job has been started')
        self._prejobs = tuple(prejobs)

    def presignal(self, job, signal):
        """
        Do not :attr:`enable <is_enabled>` this job until `job` :attr:`emits
        <signal>` `signal`

        When `job` emits `signal`, this job is enabled internally (so it can be
        :meth:`started <start>`) and the `refresh_job_list` :attr:`signal` is
        emitted.

        The payload of `signal` is discarded. Register another callback for
        `signal` if it is needed.

        See also :attr:`is_enabled`.
        """
        if self.is_started:
            raise RuntimeError('Cannot set presignal after job has been started')

        # Indicate we are waiting for signal
        self.presignals[job][signal] = False

        def presignal_handler(*args, **kwargs):
            self.presignals[job][signal] = True
            # If all presignals were emitted from the expected `job`, we tell
            # the TUI to check if the we are now ready to get started.
            if all(self.presignals[job].values()):
                self.signal.emit('refresh_job_list')

        job.signal.register(signal, presignal_handler)

    @functools.cached_property
    def presignals(self):
        return collections.defaultdict(lambda: {})

    @property
    def is_enabled(self):
        """
        Whether this job is allowed to :meth:`start`

        A job is enabled if :attr:`precondition` returns `True`, all
        :attr:`prejobs` are either finished or disabled and all
        :func:`presignal`\\ s have been emitted.

        This property must be checked by the UI every time any job finishes and
        when the `refresh_job_list` :attr:`signal` is emitted. If this property
        is `True`, this job must be :meth:`started <start>` and displayed
        (unless it is :attr:`hidden`).
        """
        return (
            # Prejobs
            all(
                (prejob.is_finished or not prejob.is_enabled)
                for prejob in self.prejobs
            )
            # Presignals
            and all(
                was_emitted
                for job_, presignals in self.presignals.items()
                for signal_, was_emitted in presignals.items()
            )
            # Precondition
            and self.precondition()
        )

    @property
    def kwargs(self):
        """Keyword arguments from instantiation as :class:`dict`"""
        return self._kwargs

    @property
    def signal(self):
        """
        :class:`~.signal.Signal` instance

        The following signals are added by the base class. Subclasses can add
        their own signals.

        ``executed``
            Emitted after :meth:`execute` was called. When output is read from
            cache, this signal is not emitted. Registered callbacks get the job
            instance as a positional argument.

        ``finished``
            Emitted when :meth:`finish` is called or when output is read from
            cache. Registered callbacks get the job instance as a positional
            argument.

        ``output``
            Emitted when :meth:`send` is called or when output is read from
            cache. Registered callbacks get the value passed to :meth:`send` as
            a positional argument.

        ``info``
            Emitted when :attr:`info` is set. Registered callbacks get the new
            :attr:`info` as a positional argument.

        ``warning``
            Emitted when :meth:`warn` is called. Registered callbacks get the
            value passed to :meth:`warn` as a positional argument.

        ``error``
            Emitted when :meth:`error` is called. Registered callbacks get the
            value passed to :meth:`error` as a positional argument.

        ``refresh_job_list``
            Emitted when the UI should to refresh the job list. The UI usually
            only does this when a job finishes. This signal allows a job to
            force the refresh immediately. Registered callbacks get no
            arguments.
        """
        return self._signal

    def __init__(self, *, home_directory=None, cache_directory=None,
                 ignore_cache=False, no_output_is_ok=False, hidden=False, autostart=True,
                 precondition=None, prejobs=(), presignals={}, callbacks={},
                 **kwargs):
        self._home_directory = home_directory if home_directory else ''
        self._cache_directory = cache_directory if cache_directory else constants.DEFAULT_CACHE_DIRECTORY
        self._ignore_cache = bool(ignore_cache)
        self._no_output_is_ok = bool(no_output_is_ok)
        self._hidden = bool(hidden)
        self._autostart = bool(autostart)
        self._is_started = False
        self._is_executed = False
        self._exception = None
        self._output = []
        self._warnings = []
        self._errors = []
        self._info = ''
        self._tasks = []
        self._finished_event = asyncio.Event()
        self.precondition = precondition if precondition is not None else (lambda: True)
        self.prejobs = prejobs

        self._signal = utils.signal.Signal(
            'output',
            'info',
            'warning',
            'error',
            'executed',
            'finished',
            'refresh_job_list',
        )
        self._signal.register('output', lambda output: self._output.append(str(output)))
        self._signal.register('warning', lambda warning: self._warnings.append(warning))
        self._signal.register('error', lambda error: self._errors.append(error))
        self._signal.record('output')

        self._kwargs = kwargs
        self.initialize(**kwargs)

        # Add signal callbacks after `initialize` had the chance to add custom
        # signals.
        for signal_name, callback in callbacks.items():
            if callable(callback):
                self._signal.register(signal_name, callback)
            else:
                for cb in callback:
                    self._signal.register(signal_name, cb)

        # Register required signals from other jobs
        for job, signals in presignals.items():
            for signal in signals:
                self.presignal(job, signal)

    def initialize(self):
        """
        Called by :meth:`__init__` with additional keyword arguments

        This method should handle its arguments and return quickly.
        """

    def execute(self):
        """
        Do the job, e.g. ask for user input or start a task or subprocess

        This method should not be called directly. Instead, :meth:`start` should
        be called to benefit from caching, to make sure all conditions are met,
        etc.

        This method must not block. It should return as quickly as possible.
        """

    def start(self):
        """
        Called by the main entry point when this job is executed

        If there is cached output available, load it and call :meth:`finish`.
        Otherwise, call :meth:`execute`.

        :raise RuntimeError: if this method is called multiple times or if
            reading from cache file fails unexpectedly
        """
        if self.is_finished:
            # This happens if another jobs fails before this job is started and
            # the job manager (e.g. tui.TUI) finishes all jobs before it handles
            # the error from the failed job.
            _log.debug('%s: Not executing already finished job', self.name)
        elif not self.is_enabled:
            # Job is waiting for prejob, presignal, precondition, etc
            _log.debug('%s: Not executing disabled job', self.name)
            # _log.debug(
            #     '%s: Not executing disabled job: prejobs=%s, presignals=%s, precondition=%s',
            #     self.name,
            #     [f'{prejob.name}:' + 'finished' if prejob.is_finished else 'running' for prejob in self.prejobs],
            #     self._presignals,
            #     self.precondition(),
            # )
        else:
            if self._is_started:
                raise RuntimeError('start() was already called')
            else:
                self._is_started = True

            try:
                cache_was_read = self._read_cache()
            except BaseException:
                self.finish()
                raise

            if cache_was_read:
                _log.debug('%s: Using cached output', self.name)
                self.finish()
            else:
                _log.debug('%s: Executing', self.name)
                self._is_executed = True
                self.execute()
                self.signal.emit('executed', self)

    @property
    def is_started(self):
        """Whether :meth:`start` was called while this job :attr:`is_enabled`"""
        return self._is_started

    async def wait(self):
        """
        Wait for this job to finish

        Subclasses that need to wait for I/O should do so by overriding this
        method.

        Subclasses must call the parent's method (``super().wait()``).

        This method returns when :meth:`finish` is called.

        :attr:`is_finished` is `False` before this method returns and `True`
        afterwards.

        Calling this method multiple times simultaneously is safe.

        :raise: The first exception given to :meth:`exception`
        """
        await self._finished_event.wait()
        if self.raised:
            raise self.raised

    def finish(self):
        """
        Cancel this job if it is not finished yet and emit ``finished`` signal

        Calling this method unblocks any calls to :meth:`wait`.

        :attr:`is_finished` is `True` after this method was called.

        Calling this method from multiple coroutines simultaneously is safe.

        Subclasses must call the parent's method (``super().finish()``).

        This method does not block.
        """
        if not self.is_finished:

            # Do not cancel tasks for now because it can print
            # RuntimeWarning: coroutine ... was never awaited
            #
            # to stderr, for example when release-name command is already
            # cached:
            # $ upsies rn ...
            # $ upsies rn ...  # Same command
            #
            # This may cause problems if long tasks are running when we call
            # finish().

            # for task in self._tasks:
            #     if not task.done():
            #         _log.debug('%s: Cancelling %r', self.name, task)
            #         task.cancel()

            self._finished_event.set()
            self.signal.emit('finished', self)
            self._write_cache()

    @property
    def is_finished(self):
        """Whether :meth:`finish` was called"""
        return self._finished_event.is_set()

    @property
    def exit_code(self):
        """`0` if job was successful, ``> 0`` otherwise, `None` while job is not finished"""
        if self.is_finished:
            if self.errors or self.raised:
                return 1
            elif not self.output and not self.no_output_is_ok:
                return 1
            else:
                return 0

    def send(self, output):
        """
        Append `output` to :attr:`output` and emit ``output`` signal

        .. note:: All output is  converted to :class:`str` because it is cached
                  and pickling adds a lot of complexity that is usually not needed.
        """
        if not self.is_finished:
            self.signal.emit('output', str(output))

    @property
    def output(self):
        """Immutable sequence of strings passed to :meth:`send`"""
        return tuple(self._output)

    @property
    def info(self):
        """
        String that is only displayed while the job is running and not part of the
        job's output

        Setting this property emits the ``info`` signal.
        """
        return self._info

    @info.setter
    def info(self, info):
        self._info = str(info)
        self.signal.emit('info', info)

    def warn(self, warning):
        """Append `warning` to :attr:`warnings` and emit ``warning`` signal"""
        if not self.is_finished:
            self.signal.emit('warning', str(warning))

    @property
    def warnings(self):
        """
        Sequence of non-critical error messages the user can override or resolve

        Unlike :attr:`errors`, warnings do not imply failure by default.
        """
        return tuple(self._warnings)

    def clear_warnings(self):
        """Empty :attr:`warnings`"""
        if not self.is_finished:
            self._warnings.clear()

    def error(self, error, finish=True):
        """
        Append `error` to :attr:`errors`, emit ``error`` signal and :meth:`finish`
        this job

        Do nothing if this job is already finished.

        :param bool finish: Whether to :meth:`finish` this job

            .. note:: This should only be used to report multiple errors before
                      finishing the job manually.
        """
        if not self.is_finished:
            self.signal.emit('error', error)
            if finish:
                self.finish()

    @property
    def errors(self):
        """
        Sequence of critical errors (strings or exceptions)

        By default, :attr:`exit_code` is non-zero if any errors were reported.
        """
        return tuple(self._errors)

    def exception(self, exception):
        """
        Make `exception` available as :attr:`raised` and call :meth:`finish`

        .. warning:: Setting an exception means you want to throw a traceback in
                     the user's face.

        :param Exception exception: Exception instance
        """
        import traceback
        tb = ''.join(traceback.format_exception(
            type(exception), exception, exception.__traceback__))
        _log.debug('Exception in %s: %s', self.name, tb)
        self._exception = exception
        self.finish()

    @property
    def raised(self):
        """Exception passed to :meth:`exception`"""
        return self._exception

    def attach_task(self, coro, callback=None, finish_when_done=False):
        """
        Run asynchronous coroutine in background task and return immediately

        This method is a wrapper around :func:`~.utils.run_task` that finishes
        the job when `coro` returns or raises.

        Any exceptions from `coro` are passed to :meth:`exception`.

        `coro` is automatically and silently cancelled by :meth:`finish`.

        :param coro: Any awaitable object
        :param callback: Callable that is called with the return value of `coro`
        :param bool finish_when_done: Whether to call :meth:`finish` when `coro`
            returns or raises an exception

        :return: :class:`asyncio.Task` instance
        """
        if self.is_finished:
            raise RuntimeError(f'Job is already finished: {self.name}')

        def callback_(result):
            try:
                if isinstance(result, BaseException):
                    _log.debug('%s: Handling exception from %r: %r', self.name, coro, result)
                    self.exception(result)
                else:
                    if callback:
                        try:
                            callback(result)
                        except BaseException as e:
                            _log.debug('%s: Handling exception from callback %r: %r', self.name, callback, result)
                            self.exception(e)
                    return result
            finally:
                if finish_when_done:
                    self.finish()

        task = utils.run_task(coro, callback=callback_)
        self._tasks.append(task)
        return task

    async def await_tasks(self):
        """Block until all coroutines passed to :meth:`attach_task` are done"""
        for task in self._tasks:
            await task

    def _write_cache(self):
        """
        Store recorded signals in :attr:`cache_file`

        Emitted signals are serialized with :meth:`_serialize_for_cache`.

        :raise RuntimeError: if writing :attr:`cache_file` fails
        """
        if self.signal.emissions and self.exit_code == 0 and self.cache_file and self._is_executed:
            emissions_serialized = self._serialize_for_cache(self.signal.emissions)
            _log.debug('%s: Caching emitted signals: %r', self.name, self.signal.emissions)
            try:
                with open(self.cache_file, 'wb') as f:
                    f.write(emissions_serialized)
                    f.write(b'\n')
            except OSError as e:
                msg = e.strerror if e.strerror else str(e)
                raise RuntimeError(f'Unable to write cache {self.cache_file}: {msg}')

    def _read_cache(self):
        """
        Read cached :attr:`~.signal.Signal.emissions` from :attr:`cache_file`

        Emitted signals are deserialized with :meth:`_deserialize_from_cache`.

        :raise RuntimeError: if :attr:`cache_file` exists and is unreadable

        :return: `True` if cache file was read, `False` otherwise
        """
        if not self._ignore_cache and self.cache_file and os.path.exists(self.cache_file):
            try:
                with open(self.cache_file, 'rb') as f:
                    emissions_serialized = f.read()
            except OSError as e:
                msg = e.strerror if e.strerror else str(e)
                raise RuntimeError(f'Unable to read cache {self.cache_file}: {msg}')
            else:
                emissions_deserialized = self._deserialize_from_cache(emissions_serialized)
                _log.debug('%s: Replaying cached signals: %r', self.name, emissions_deserialized)
                if emissions_deserialized:
                    self.signal.replay(emissions_deserialized)
                    return True
        return False

    def _serialize_for_cache(self, emissions):
        """
        Convert emitted signals to cache format

        :param emissions: See :attr:`Signal.emissions`

        :return: :class:`bytes`
        """
        return pickle.dumps(emissions, protocol=0, fix_imports=False)

    def _deserialize_from_cache(self, emissions_serialized):
        """
        Convert return value of :meth:`_serialize_for_cache` back to emitted signals

        :param emissions_serialized: :class:`bytes` object

        :return: See :attr:`Signal.emissions`
        """
        return pickle.loads(emissions_serialized)

    _max_filename_len = 255

    @functools.cached_property
    def cache_file(self):
        """
        File path in :attr:`cache_directory` to store cached :attr:`output` in

        If this property returns `None`, cache is not read or written.
        """
        cache_id = self.cache_id
        if cache_id is None:
            return None
        elif not cache_id:
            filename = f'{self.name}.out'
        else:
            # Avoid file name being too long. 255 bytes seems common.
            # https://en.wikipedia.org/wiki/Comparison_of_file_systems#Limits
            max_len = self._max_filename_len - len(self.name) - len('..out')
            cache_id_str = self._cache_id_as_string(cache_id)
            if len(cache_id_str) > max_len:
                cache_id_str = ''.join((
                    cache_id_str[:int(max_len / 2 - 1)],
                    '…',
                    cache_id_str[-int(max_len / 2 - 1):],
                ))
            filename = f'{self.name}.{cache_id_str}.out'
            filename = utils.fs.sanitize_filename(filename)
        return os.path.join(self.cache_directory, filename)

    @functools.cached_property
    def cache_id(self):
        """
        Unique object based on the job's input data

        The return value is turned into a string. If it is a non-string sequence
        or a mapping, items are converted to strings and joined with ",".
        Multibyte characters and directory delimiters are replaced.

        If this property returns `None`, :attr:`cache_file` is not read or
        written.

        If this property returns any other falsy value, :attr:`name` is used.
        """
        return ''

    def _cache_id_as_string(self, value):
        if isinstance(value, collections.abc.Mapping):
            return ','.join((f'{self._cache_id_value_as_string(k)}={self._cache_id_as_string(v)}'
                             for k, v in value.items()))
        elif isinstance(value, collections.abc.Iterable) and not isinstance(value, str):
            return ','.join((self._cache_id_as_string(v) for v in value))
        elif isinstance(value, (str, os.PathLike)) and os.path.exists(value):
            # Use same cache file for absolute and relative paths
            return str(os.path.realpath(value))
        else:
            return self._cache_id_value_as_string(value)

    _object_without_str_regex = re.compile(r'^<.*>$')

    def _cache_id_value_as_string(self, value):
        # Avoid multibyte characters stay below maximum file length
        value_string = unidecode.unidecode(str(value))
        # Check if `value` has a proper string representation to prevent random
        # cache IDs. We don't want "<foo.bar object at 0x...>" in our cache ID.
        if self._object_without_str_regex.search(value_string):
            raise RuntimeError(f'{type(value)!r} has no string representation')
        return value_string


class QueueJobBase(JobBase):
    """
    Subclass of :class:`JobBase` with an :class:`asyncio.Queue`

    This job is used to process input asynchronously, e.g. from another job. For
    example, :meth:`enqueue` from :class:`~.jobs.imghost.ImageHostJob` can be
    connected to the ``output`` :class:`~.utils.signal.Signal` of
    :class:`~.jobs.screenshots.ScreenshotsJob`.

    It's also possible to use this job conventionally by passing a sequence of
    values as the `enqueue` argument. This processes all values and finishes the
    job without waiting for more.

    The :meth:`initialize` method of subclasses must accept `enqueue` as a
    keyword argument.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._queue = asyncio.Queue()
        self._read_queue_task = None
        self._enqueue_args = kwargs.get('enqueue', ())

    def execute(self):
        self._read_queue_task = asyncio.ensure_future(self._read_queue())
        if self._enqueue_args:
            for value in self._enqueue_args:
                self.enqueue(value)
            self.finalize()

    async def _read_queue(self):
        while True:
            value = await self._queue.get()
            if value is None or self.is_finished:
                break
            else:
                try:
                    await self.handle_input(value)
                except asyncio.CancelledError:
                    _log.debug('%s: Job was cancelled while handling %r', self.name, value)
                    break
                except BaseException as e:
                    self.exception(e)
                    break
        self.finish()

    @abc.abstractmethod
    async def handle_input(self, value):
        """Handle `value` from queue"""

    def enqueue(self, value):
        """Put `value` in queue"""
        self._queue.put_nowait(value)

    def finalize(self):
        """Finish after all currently queued values are handled"""
        self._queue.put_nowait(None)

    def finish(self):
        """
        Stop reading from queue

        Unlike :meth:`JobBase.finish`, this does not finish the job. The job is
        not finished until :meth:`wait` returns.
        """
        if self._read_queue_task:
            self._read_queue_task.cancel()
        super().finish()

    async def wait(self):
        """Wait for internal queue reading task and finish job"""
        if self._read_queue_task:
            try:
                await self._read_queue_task
            except asyncio.CancelledError:
                pass
        await super().wait()



"""
Video metadata
"""

import collections
import functools
import glob
import json
import os
import re

from .. import constants, errors, utils
from . import closest_number, fs, os_family, release, subproc

import logging  # isort:skip
_log = logging.getLogger(__name__)

natsort = utils.LazyModule(module='natsort', namespace=globals())


NO_DEFAULT_VALUE = object()


if os_family() == 'windows':
    _mediainfo_executable = 'mediainfo.exe'
    _ffprobe_executable = 'ffprobe.exe'
else:
    _mediainfo_executable = 'mediainfo'
    _ffprobe_executable = 'ffprobe'


@functools.lru_cache(maxsize=None)
def _run_mediainfo(video_file_path, *args):
    fs.assert_file_readable(video_file_path)
    cmd = (_mediainfo_executable, video_file_path) + args

    # Translate DependencyError to ContentError so callers have to expect less
    # exceptions. Do not catch ProcessError because things like wrong mediainfo
    # arguments are bugs.
    try:
        return subproc.run(cmd, cache=True)
    except errors.DependencyError as e:
        raise errors.ContentError(e)


@functools.lru_cache(maxsize=None)
def mediainfo(path, only_first=True):
    """
    ``mediainfo`` output as a string

    The parent directory of `path` is redacted.

    :param str path: Path to video file or directory
    :param bool only_first: Whether to return mediainfo for each video file as a
        :class:`list` or to return mediainfo of the first video file (see
        :func:`find_videos`)

    :raise ContentError: if anything goes wrong
    """
    def remove_parent_directory(mi, parent_directory=utils.fs.dirname(path)):
        if parent_directory:
            return mi.replace(parent_directory + os.sep, '')
        else:
            return mi

    unique_id_default = 'Unique ID                                : 0 (0x0)'
    unique_id_regex = re.compile(r'^Unique ID\s*:\s*\d+', flags=re.MULTILINE)

    def ensure_unique_id_exists(mi):
        # Some mediainfo parsers expect a unique ID, but it doesn't always exist.
        if unique_id_regex.search(mi):
            return mi
        else:
            assert mi.startswith('General\n')
            return (
                mi[:len('General\n')]
                + unique_id_default + '\n'
                + mi[len('General\n'):]
            )

    def apply_fixes(mi):
        return remove_parent_directory(
            ensure_unique_id_exists(mi)
        )

    if not os.path.isdir(path):
        # If we get a file path, don't complain if it's not a video file. This
        # allows us to get mediainfo for .IFO files even though they are not
        # videos.
        mediainfo = _run_mediainfo(path)
        return apply_fixes(mediainfo)

    elif only_first:
        video_file = find_videos(path)[0]
        mediainfo = _run_mediainfo(video_file)
        return apply_fixes(mediainfo)

    else:
        video_files = find_videos(path)
        mediainfos = []
        for video_file in video_files:
            mediainfo = _run_mediainfo(video_file)
            mediainfos.append(apply_fixes(mediainfo))
        return mediainfos


@functools.lru_cache(maxsize=None)
def duration(path, default=NO_DEFAULT_VALUE):
    """
    Return video duration in seconds (float) or ``0.0`` if it can't be
    determined

    :param str path: Path to video file or directory

        For directories, the first video is used. See :func:`find_videos`.
    :param default: Return value if `path` doesn't exist, raise
        :exc:`~.ContentError` if not provided

    :raise ContentError: if anything goes wrong
    """
    if not os.path.exists(path) and default is not NO_DEFAULT_VALUE:
        return default
    return _duration(find_videos(path)[0])

def _duration(video_file_path):
    try:
        return _duration_from_ffprobe(video_file_path)
    except (RuntimeError, errors.DependencyError, errors.ProcessError):
        return _duration_from_mediainfo(video_file_path)

def _duration_from_ffprobe(video_file_path):
    cmd = (
        _ffprobe_executable,
        '-v', 'error', '-show_entries', 'format=duration',
        '-of', 'default=noprint_wrappers=1:nokey=1',
        make_ffmpeg_input(video_file_path),
    )
    length = subproc.run(cmd, ignore_errors=True)
    try:
        return float(length.strip())
    except ValueError:
        raise RuntimeError(f'Unexpected output from {cmd}: {length!r}')

def _duration_from_mediainfo(video_file_path):
    tracks = _tracks(video_file_path)
    try:
        return float(tracks['General'][0]['Duration'])
    except (KeyError, IndexError, TypeError, ValueError):
        raise RuntimeError(f'Unexpected tracks from {video_file_path}: {tracks!r}')


@functools.lru_cache(maxsize=None)
def tracks(path, default=NO_DEFAULT_VALUE):
    """
    ``mediainfo --Output=JSON`` as dictionary that maps each track's ``@type``
    to a list of the tracks

    :param str path: Path to video file or directory

        For directories, the first video is used. See :func:`find_videos`.
    :param default: Return value if `path` doesn't exist, raise
        :exc:`~.ContentError` if not provided

    :raise ContentError: if anything goes wrong
    """
    if default is not NO_DEFAULT_VALUE and not os.path.exists(path):
        return default

    try:
        main_videos = find_videos(path)
    except errors.ContentError:
        # No video file found - use path as is. This allows us to get tracks
        # from an .IFO file, which isn't a video, so find_videos() raises
        # ContentError. _tracks()/_run_mediainfo() will still raise ContentError
        # if `path` mediainfo can't handle it.
        video_filepath = path
    else:
        video_filepath = main_videos[0]

    return _tracks(video_filepath)

def _tracks(video_file_path):
    stdout = _run_mediainfo(video_file_path, '--Output=JSON')
    tracks = {}
    try:
        for track in json.loads(stdout)['media']['track']:
            if track['@type'] not in tracks:
                tracks[track['@type']] = []
            tracks[track['@type']].append(track)
        return tracks
    except (ValueError, TypeError) as e:
        raise RuntimeError(f'{video_file_path}: Unexpected mediainfo output: {stdout}: {e}')
    except KeyError as e:
        raise RuntimeError(f'{video_file_path}: Unexpected mediainfo output: {stdout}: Missing field: {e}')


@functools.lru_cache(maxsize=None)
def default_track(type, path, default=NO_DEFAULT_VALUE):
    """
    Return default track

    :param str type: "video", "audio" or "text"
    :param str path: Path to video file or directory

        For directories, the first video is used. See :func:`find_videos`.
    :param default: Return value if `path` doesn't exist or doesn't have a
        `type` track

         If this is not provided, raise :exc:`~.ContentError`.

    :raise ContentError: if anything goes wrong and `default` is not provided
    """
    if default is not NO_DEFAULT_VALUE and not os.path.exists(path):
        return default

    all_tracks = tracks(path)

    # Find track marked as default
    try:
        for track in all_tracks[type.capitalize()]:
            if track.get('Default') == 'Yes':
                return track
    except KeyError:
        pass

    # Default to first track
    try:
        return all_tracks[type.capitalize()][0]
    except (KeyError, IndexError):
        pass

    if default is NO_DEFAULT_VALUE:
        raise errors.ContentError(f'{path}: No {type.lower()} track found: {all_tracks!r}')
    else:
        return default


@functools.lru_cache(maxsize=None)
def width(path, default=NO_DEFAULT_VALUE):
    """
    Return displayed width of video file `path`

    :param str path: Path to video file or directory

        For directories, the first video is used. See :func:`find_videos`.
    :param default: Return value if `path` doesn't exist, raise
        :exc:`~.ContentError` if not provided

    :raise ContentError: if width can't be determined
    """
    if default is not NO_DEFAULT_VALUE and not os.path.exists(path):
        return default

    video_track = default_track('video', path)
    return _get_display_width(video_track)

def _get_display_width(video_track):
    try:
        width = int(video_track['Width'])
    except (KeyError, ValueError, TypeError):
        raise errors.ContentError('Unable to determine video width')
    else:
        _log.debug('Stored width: %r', width)
        # Actual aspect ratio may differ from display aspect ratio,
        # e.g. 960 x 534 is scaled up to 1280 x 534
        par = float(video_track.get('PixelAspectRatio') or 1.0)
        if par > 1.0:
            _log.debug('Display width: %r * %r = %r', width, par, width * par)
            width = width * par
        return int(width)


@functools.lru_cache(maxsize=None)
def height(path, default=NO_DEFAULT_VALUE):
    """
    Return displayed height of video file `path`

    :param str path: Path to video file or directory

        For directories, the first video is used. See :func:`find_videos`.
    :param default: Return value if `path` doesn't exist, raise
        :exc:`~.ContentError` if not provided

    :raise ContentError: if height can't be determined
    """
    if default is not NO_DEFAULT_VALUE and not os.path.exists(path):
        return default

    video_track = default_track('video', path)
    return _get_display_height(video_track)

def _get_display_height(video_track):
    try:
        height = int(video_track['Height'])
    except (KeyError, ValueError, TypeError):
        raise errors.ContentError('Unable to determine video height')
    else:
        _log.debug('Stored height: %r', height)
        # Actual aspect ratio may differ from display aspect ratio,
        # e.g. 960 x 534 is scaled up to 1280 x 534
        par = float(video_track.get('PixelAspectRatio') or 1.0)
        if par < 1.0:
            _log.debug('Display height: (1 / %r) * %r = %r', par, height, (1 / par) * height)
            height = (1 / par) * height
        return int(height)


@functools.lru_cache(maxsize=None)
def resolution(path, default=NO_DEFAULT_VALUE):
    """
    Return resolution of video file `path` as :class:`str` (e.g. "1080p")

    :param str path: Path to video file or directory

        For directories, the first video is used. See :func:`find_videos`.
    :param default: Return value if `path` doesn't exist, raise
        :exc:`~.ContentError` if not provided

    :raise ContentError: if resolution can't be determined
    """
    if default is not NO_DEFAULT_VALUE and not os.path.exists(path):
        return default

    video_track = default_track('video', path)
    std_resolution = _get_closest_standard_resolution(video_track)
    scan_type = _scan_type(video_track)
    return f'{std_resolution}{scan_type}'


@functools.lru_cache(maxsize=None)
def resolution_int(path, default=NO_DEFAULT_VALUE):
    """
    Return resolution of video file `path` as :class:`int` (e.g. ``1080``)

    :param str path: Path to video file or directory

        For directories, the first video is used. See :func:`find_videos`.
    :param default: Return value if `path` doesn't exist, raise
        :exc:`~.ContentError` if not provided

    :raise ContentError: if resolution can't be determined
    """
    if default is not NO_DEFAULT_VALUE and not os.path.exists(path):
        return default

    video_track = default_track('video', path)
    std_resolution = _get_closest_standard_resolution(video_track)
    return std_resolution


def _scan_type(video_track):
    # "p" or "i", default to "p"
    scan_type = str(video_track.get('ScanType', 'Progressive')).lower()
    if scan_type in ('interlaced', 'mbaff', 'paff'):
        return 'i'
    else:
        return 'p'

_standard_resolutions = (
    # (width, height, standard resolution)
    # 4:3
    (640, 480, 480),
    (720, 540, 540),
    (768, 576, 576),
    (960, 720, 720),
    (1440, 1080, 1080),
    (2880, 2160, 2160),
    (5760, 4320, 4320),

    # 16:9
    (853, 480, 480),
    (960, 540, 540),
    (1024, 576, 576),
    (1280, 720, 720),
    (1920, 1080, 1080),
    (3840, 2160, 2160),
    (7680, 4320, 4320),

    # 21:9
    (853, 365, 480),
    (960, 411, 540),
    (1024, 438, 576),
    (1280, 548, 720),
    (1920, 822, 1080),
    (3840, 1645, 2160),
    (7680, 3291, 4320),
)

def _get_closest_standard_resolution(video_track):
    actual_width = _get_display_width(video_track)
    actual_height = _get_display_height(video_track)

    # Find distances from actual display width/height to each standard
    # width/height. Categorize them by standard ratio.
    distances = collections.defaultdict(lambda: {})
    for std_width, std_height, std_resolution in _standard_resolutions:
        std_aspect_ratio = round(std_width / std_height, 1)
        w_dist = abs(std_width - actual_width)
        h_dist = abs(std_height - actual_height)
        resolution = (std_width, std_height, std_resolution)
        distances[std_aspect_ratio][w_dist] = distances[std_aspect_ratio][h_dist] = resolution

    # Find standard aspect ratio closest to the given aspect ratio
    actual_aspect_ratio = round(actual_width / actual_height, 1)
    std_aspect_ratios = tuple(distances)
    closest_std_aspect_ratio = closest_number(actual_aspect_ratio, std_aspect_ratios)

    # Pick the standard resolution with the lowest distance to the given resolution
    dists = distances[closest_std_aspect_ratio]
    std_width, std_height, std_resolution = sorted(dists.items())[0][1]

    _log.debug('Closest standard resolution: %r x %r [%.1f] -> %r x %r [%.1f] -> %r',
               actual_width, actual_height, actual_aspect_ratio,
               std_width, std_height, closest_std_aspect_ratio,
               std_resolution)
    return std_resolution


@functools.lru_cache(maxsize=None)
def frame_rate(path, default=NO_DEFAULT_VALUE):
    """
    Return FPS of default video track or ``0.0`` if it can't be determined

    :param str path: Path to video file or directory

        For directories, the first video is used. See :func:`find_videos`.
    :param default: Return value if `path` doesn't exist, raise
        :exc:`~.ContentError` if not provided

    :raise ContentError: if no video track is found
    """
    if default is not NO_DEFAULT_VALUE and not os.path.exists(path):
        return default

    video_track = default_track('video', path)
    try:
        return float(video_track['FrameRate'])
    except (KeyError, ValueError, TypeError):
        return 0.0


@functools.lru_cache(maxsize=None)
def bit_depth(path, default=NO_DEFAULT_VALUE):
    """
    Return bit depth of default video track or ``0`` if it can't be determined

    :param str path: Path to video file or directory

        For directories, the first video is used. See :func:`find_videos`.
    :param default: Return value if `path` doesn't exist, raise
        :exc:`~.ContentError` if not provided

    :raise ContentError: if no video track is found
    """
    if default is not NO_DEFAULT_VALUE and not os.path.exists(path):
        return default

    video_track = default_track('video', path)
    try:
        return int(video_track['BitDepth'])
    except (KeyError, ValueError):
        return 0


hdr_formats = {
    'DV': {
        # Dolby Vision[ / <more information>]
        'HDR_Format': re.compile(r'^Dolby Vision'),
    },
    # "HDR10+ Profile A" or "HDR10+ Profile B"
    'HDR10+': {
        'HDR_Format_Compatibility': re.compile(r'HDR10\+'),
    },
    'HDR10': (
        {'HDR_Format_Compatibility': re.compile(r'HDR10(?!\+)')},
        {'colour_primaries': re.compile(r'BT\.2020')},
    ),
    'HDR': (
        {'HDR_Format_Compatibility': re.compile(r'HDR')},
        {'HDR_Format': re.compile(r'HDR')},
    ),
}
"""
Map HDR format names (e.g. "Dolby Vision") to dictionaries that map
mediainfo video track fields to regular expressions that must match the fields'
values
"""

@functools.lru_cache(maxsize=None)
def hdr_format(path, default=NO_DEFAULT_VALUE):
    """
    Return HDR format based on `hdr_formats`, e.g. "HDR10", or empty string

    :param str path: Path to video file or directory

        For directories, the first video is used. See :func:`find_videos`.
    :param default: Return value if `path` doesn't exist, raise
        :exc:`~.ContentError` if not provided

    :raise ContentError: if no video track is found
    """
    if default is not NO_DEFAULT_VALUE and not os.path.exists(path):
        return default

    def is_match(video_track, conditions):
        if isinstance(conditions, collections.abc.Mapping):
            # `conditions` maps field names to field value matching regexes
            for key, regex in conditions.items():
                value = video_track.get(key, '')
                if regex.search(value):
                    return True
        else:
            # `conditions` is a sequence of mappings that map field names to
            # field value matching regexes (see above)
            for condition in conditions:
                if is_match(video_track, condition):
                    return True
        return False

    video_track = default_track('video', path)
    for hdr_format, conditions in hdr_formats.items():
        if is_match(video_track, conditions):
            return hdr_format
    return ''


@functools.lru_cache(maxsize=None)
def has_dual_audio(path, default=NO_DEFAULT_VALUE):
    """
    Return `True` if `path` contains multiple audio tracks with different
    languages and one of them is English, `False` otherwise

    :param str path: Path to video file or directory

        For directories, the first video is used. See :func:`find_videos`.
    :param default: Return value if `path` doesn't exist, raise
        :exc:`~.ContentError` if not provided

    :raise ContentError: if reading `path` fails
    """
    if default is not NO_DEFAULT_VALUE and not os.path.exists(path):
        return default

    languages = set()
    audio_tracks = tracks(path).get('Audio', ())
    for track in audio_tracks:
        if 'commentary' not in track.get('Title', '').lower():
            language = track.get('Language', '')
            if language:
                languages.add(language.casefold()[:2])
    if len(languages) > 1:
        return True
    else:
        return False


def is_bt601(path, default=NO_DEFAULT_VALUE):
    """
    Whether `path` is BT.601 (~SD) video

    :param str path: Path to video file or directory

        For directories, the first video is used. See :func:`find_videos`.
    :param default: Return value if `path` doesn't exist, raise
        :exc:`~.ContentError` if not provided

    :raise ContentError: if reading `path` fails
    """
    if default is not NO_DEFAULT_VALUE and not os.path.exists(path):
        return default

    # https://rendezvois.github.io/video/screenshots/programs-choices/#color-matrix
    if _get_color_matrix(path, 'BT.601'):
        return True
    elif _get_color_matrix(path, 'BT.470 System B/G'):
        return True
    else:
        # Assume BT.601 if there is no HD video track
        video_tracks = tracks(path).get('Video', ())
        for video_track in video_tracks:
            resolution = _get_closest_standard_resolution(video_track)
            if resolution >= 720:
                return False
        return True

def is_bt709(path, default=NO_DEFAULT_VALUE):
    """
    Whether `path` is BT.709 (~HD) video

    See :func:`is_bt601`.
    """
    return _get_color_matrix(path, 'BT.709', default=default)

def is_bt2020(path, default=NO_DEFAULT_VALUE):
    """
    Whether `path` is BT.2020 (~UHD) video

    See :func:`is_bt601`.
    """
    return _get_color_matrix(path, 'BT.2020', default=default)

def _get_color_matrix(path, name, default=NO_DEFAULT_VALUE):
    if default is not NO_DEFAULT_VALUE and not os.path.exists(path):
        return default

    # https://rendezvois.github.io/video/screenshots/programs-choices/#color-matrix
    video_tracks = tracks(path).get('Video', ())
    for video_track in video_tracks:
        if video_track.get('matrix_coefficients', '').startswith(name):
            return True
    return False


@functools.lru_cache(maxsize=None)
def languages(path, default=NO_DEFAULT_VALUE, exclude_commentary=True):
    """
    Extract ``Language`` fields from each track for each track type

    Example return value:

    .. code::

        {
            'Audio': {'en', 'dk'},
            'Text': {'en', 'cz', 'pt'},
        }

    :param str path: Path to video file or directory
    :param default: Default language for each track that doesn't specify one
    :param exclude_commentary: Ignore any track with a ``Title`` field that
        contains the string "commentary" (case-insensitive)

    :raise ContentError: if reading `path` fails
    """
    all_tracks = utils.video.tracks(path)
    languages = collections.defaultdict(set)
    for type, tracks in all_tracks.items():
        for track in tracks:
            title = track.get('Title', '').lower()
            if not exclude_commentary or 'commentary' not in title:
                language = track.get('Language', None)
                if language is not None:
                    languages[type].add(language)
                elif default is not NO_DEFAULT_VALUE:
                    languages[type].add(default)
    return dict(languages)


@functools.lru_cache(maxsize=None)
def has_commentary(path, default=NO_DEFAULT_VALUE):
    """
    Return `True` if `path` has an audio track with "Commentary"
    (case-insensitive) in its title, `False` otherwise

    :param str path: Path to video file or directory

        For directories, the first video is used. See :func:`find_videos`.
    :param default: Return value if `path` doesn't exist, raise
        :exc:`~.ContentError` if not provided

    :raise ContentError: if reading `path` fails
    """
    if default is not NO_DEFAULT_VALUE and not os.path.exists(path):
        return default

    audio_tracks = tracks(path).get('Audio', ())
    for track in audio_tracks:
        title = track.get('Title', '').lower()
        if any(word in title for word in ('commentary', 'comments')):
            return True
    return False


_audio_format_translations = (
    # (<format>, <<key>:<regex> dictionary>)
    # - All <regex>s must match each <key> to identify <format>.
    # - All identified <format>s are appended (e.g. "TrueHD Atmos").
    # - {<key>: None} means <key> must not exist.
    ('AAC', {'Format': re.compile(r'^AAC$')}),
    ('DD', {'Format': re.compile(r'^AC-3$')}),
    ('DDP', {'Format': re.compile(r'^E-AC-3$')}),
    ('TrueHD', {'Format': re.compile(r'MLP ?FBA')}),
    ('TrueHD', {'Format_Commercial_IfAny': re.compile(r'TrueHD')}),
    ('Atmos', {'Format_Commercial_IfAny': re.compile(r'Atmos')}),
    ('DTS', {'Format': re.compile(r'^DTS$'), 'Format_Commercial_IfAny': None}),
    ('DTS-ES', {'Format_Commercial_IfAny': re.compile(r'DTS-ES')}),
    ('DTS-HD', {'Format_Commercial_IfAny': re.compile(r'DTS-HD(?! Master Audio)')}),
    ('DTS-HD MA', {
        'Format_Commercial_IfAny': re.compile(r'DTS-HD Master Audio'),
        'Format_AdditionalFeatures': re.compile(r'XLL$'),
    }),
    ('DTS:X', {'Format_AdditionalFeatures': re.compile(r'XLL X')}),
    ('FLAC', {'Format': re.compile(r'FLAC')}),
    ('MP3', {'Format': re.compile(r'MPEG Audio')}),
    ('Vorbis', {'Format': re.compile(r'\bVorbis\b')}),
    ('Vorbis', {'Format': re.compile(r'\bOgg\b')}),
    ('Opus', {'Format': re.compile(r'\bOpus\b')}),
)

@functools.lru_cache(maxsize=None)
def audio_format(path, default=NO_DEFAULT_VALUE):
    """
    Return audio format (e.g. "AAC", "MP3") or empty string

    :param str path: Path to video file or directory

        For directories, the first video is used. See :func:`find_videos`.
    :param default: Return value if `path` doesn't exist or has no audio track,
        raise :exc:`~.ContentError` if not provided

    :raise ContentError: if no audio track is found and no `default` is provided
    """
    if default is not NO_DEFAULT_VALUE and not os.path.exists(path):
        return default

    try:
        audio_track = default_track('audio', path)
    except errors.ContentError:
        if default is NO_DEFAULT_VALUE:
            raise
        else:
            audio_track = {}

    def is_match(regexs, audio_track):
        for key,regex in regexs.items():
            if regex is None:
                if key in audio_track:
                    # `key` must not exists but it does
                    return False
            else:
                # regex is not None
                if key not in audio_track:
                    # `key` doesn't exist
                    return False
                elif not regex.search(audio_track.get(key, '')):
                    # `key` has value that doesn't match `regex`
                    return False
        # All `regexs` match and no forbidden keys exist in `audio_track`
        return True

    parts = []
    for fmt, regexs in _audio_format_translations:
        if fmt not in parts and is_match(regexs, audio_track):
            parts.append(fmt)
    return ' '.join(parts)


# NOTE: guessit only recognizes 7.1, 5.1, 2.0 and 1.0
_audio_channels_translations = (
    ('1.0', re.compile(r'^1$')),
    ('2.0', re.compile(r'^2$')),
    ('2.0', re.compile(r'^3$')),
    ('2.0', re.compile(r'^4$')),
    ('2.0', re.compile(r'^5$')),
    ('5.1', re.compile(r'^6$')),
    ('5.1', re.compile(r'^7$')),
    ('7.1', re.compile(r'^8$')),
    ('7.1', re.compile(r'^9$')),
    ('7.1', re.compile(r'^10$')),
)

@functools.lru_cache(maxsize=None)
def audio_channels(path, default=NO_DEFAULT_VALUE):
    """
    Return audio channels (e.g. "5.1") or empty_string

    :param str path: Path to video file or directory

        For directories, the first video is used. See :func:`find_videos`.
    :param default: Return value if `path` doesn't exist or has no audio track,
        raise :exc:`~.ContentError` if not provided

    :raise ContentError: if no audio track is found and no `default` is provided
    """
    if default is not NO_DEFAULT_VALUE and not os.path.exists(path):
        return default

    try:
        audio_track = default_track('audio', path)
    except errors.ContentError:
        if default is NO_DEFAULT_VALUE:
            raise
        else:
            audio_track = {}

    audio_channels = ''
    channels = audio_track.get('Channels', '')
    for achan,regex in _audio_channels_translations:
        if regex.search(channels):
            audio_channels = achan
            break

    return audio_channels


_video_translations = (
    ('x264', {'Encoded_Library_Name': re.compile(r'^x264$')}),
    ('x265', {'Encoded_Library_Name': re.compile(r'^x265$')}),
    ('XviD', {'Encoded_Library_Name': re.compile(r'^XviD$')}),
    ('H.264', {'Format': re.compile(r'^AVC$')}),
    ('H.265', {'Format': re.compile(r'^HEVC$')}),
    ('VP9', {'Format': re.compile(r'^VP9$')}),
    ('MPEG-2', {'Format': re.compile(r'^MPEG Video$')}),
)

@functools.lru_cache(maxsize=None)
def video_format(path, default=NO_DEFAULT_VALUE):
    """
    Return video format or x264/x265/XviD if they were used or empty string if
    it can't be determined

    :param str path: Path to video file or directory

        For directories, the first video is used. See :func:`find_videos`.
    :param default: Return value if `path` doesn't exist, raise
        :exc:`~.ContentError` if not provided

    :raise ContentError: if no video track is found
    """
    if default is not NO_DEFAULT_VALUE and not os.path.exists(path):
        return default

    def translate(video_track):
        for vfmt,regexs in _video_translations:
            for key,regex in regexs.items():
                value = video_track.get(key)
                if value:
                    if regex.search(value):
                        return vfmt
        return ''

    video_track = default_track('video', path)
    video_format = translate(video_track)
    _log.debug('Detected video format: %s', video_format)
    return video_format


@functools.lru_cache(maxsize=None)
def find_videos(path):
    """
    Find video files, excluding nfo, screenshots, samples, etc

    Paths to Blu-ray trees (directories that contain a "BDMV" directory) are
    simply returned.

    If there are any "VIDEO_TS" (DVD) directories beneath `path` and its
    subdirectories, the largest VOB file for each "VIDEO_TS" directory is
    returned.

    For season packs (i.e. directories that contain files with "S..E.." in their
    name), episodes are returned and all other files are excluded.

    In all other cases that are not described above, directories are walked
    recursively, video files are detected by file extension and sorted naturally
    (e.g. "2.mkv" is sorted before "10.mkv"), and found video files are filtered
    through :func:`filter_main_videos` to exclude samples and such.

    :param str path: Path to file or directory

    :return: Sequence of video file paths

    :raise ContentError: if no video file can be found
    """
    def normalize_files(files):
        return natsort.natsorted(str(f) for f in files)

    # Blu-ray
    if os.path.isdir(os.path.join(path, 'BDMV')):
        # ffmpeg can read BDMV with the "bluray:" protocol
        return normalize_files([path])

    # DVD
    if os.path.isdir(path):
        # ffmpeg doesn't have a "dvd:" protocol so we return the largest VOB
        # file in each VIDEO_TS subdirectory beneath `path`
        files = []
        for video_ts_path in for_each_video_ts(path):
            all_vobs = utils.fs.file_list(video_ts_path, extensions=('VOB',))
            main_vob = find_main_vob(all_vobs)
            if main_vob:
                files.append(main_vob)
        if files:
            return normalize_files(files)

    # Directory containing video files or video file
    files = fs.file_list(path, extensions=constants.VIDEO_FILE_EXTENSIONS)

    # Find episodes in season pack by file name for better accuracy and to avoid
    # time consuming system calls in filter_main_videos()
    episode_regex = re.compile(
        rf'(?:^|{release.DELIM})'
        rf'(?i:S|Season{release.DELIM}?)\d+{release.DELIM}?(?i:E|Episode{release.DELIM}?)\d+'
        rf'{release.DELIM}'
    )
    episodes = [
        file
        for file in files
        if episode_regex.search(os.path.basename(file))
    ]
    if episodes:
        return normalize_files(episodes)

    if not files:
        raise errors.ContentError(f'{path}: No video file found')
    else:
        # To avoid using samples, extras, very short .VOBs (e.g. menus), remove
        # any videos that are very short compared to the average duration.
        considered_files = filter_main_videos(files)
        return normalize_files(considered_files)


def filter_main_videos(video_file_paths):
    r"""
    Filter `video_file_paths` for main video files

    Exclude any non-videos (see :const:`~.constants.VIDEO_FILE_EXTENSIONS`) and
    any videos that are smaller than 75 % of the average file size
    (e.g. samples).

    If `video_file_paths` contains any .VOB files from a DVD (path ends with
    ``/VIDEO_TS/VTS_\d+_\d+\.VOB``), return the largest .VOB file from each
    VIDEO_TS directory.

    .. note:: Because this function is decorated with
       :func:`functools.lru_cache`, `video_file_paths` should be a tuple (or any
       other hashable sequence).

    :params video_file_paths: Hashable sequence of video file paths

    :return: Filtered subset of `video_file_paths` as a sequence
    """
    # Single video file
    paths = tuple(video_file_paths)
    if len(paths) < 2:
        return paths

    # DVD (VIDEO_TS)
    main_video_ts_videos = _filter_main_videos_dvd(paths)
    if main_video_ts_videos:
        return main_video_ts_videos

    # Exclude non-video files
    only_video_paths = (
        fp for fp in paths
        if utils.fs.file_extension(fp).lower() in constants.VIDEO_FILE_EXTENSIONS
    )
    # Get size of each video
    sizes = {fp: utils.fs.file_size(fp) for fp in only_video_paths}
    # Exclude video files with no size (fs.file_size() may return `None`)
    sizes = {fp: size for fp, size in sizes.items() if size}

    # Exclude video files with size below average
    if sizes:
        avg = sum(sizes.values()) / len(sizes)
    else:
        avg = 0
    min_size = avg * 0.75
    return tuple(
        filepath for filepath, filesize in sizes.items()
        if filesize >= min_size
    )


def _filter_main_videos_dvd(video_file_paths):
    vob_sets = collections.defaultdict(lambda: [])
    # Group files by common VIDEO_TS parent directory
    for video_file_path in video_file_paths:
        match = re.search(
            rf'^(.*{utils.fs.OS_SEP_REGEX}VIDEO_TS)({utils.fs.OS_SEP_REGEX}VTS_\d+_\d+.[A-Z]+)$',
            video_file_path,
        )
        if match:
            video_ts_path = match.group(1)
            vob_sets[video_ts_path].append(video_file_path)

    # Return main VOB file for each VIDEO_TS directory
    main_vobs = []
    for video_ts_path_, vob_paths in vob_sets.items():
        main_vob = find_main_vob(vob_paths)
        if main_vob:
            main_vobs.append(main_vob)
    return main_vobs


def find_main_vobs(video_ts_files):
    """
    Return largest set of VOB files

    VOB files are split into sets by their file names, ignoring the second
    number, e.g. ``VTS_01_*.VOB``, ``VTS_02_*.VOB``, ``VTS_03_*.VOB``, etc.

    The largest set of VOB files in bytes is the main set.

    Any other files (e.g. ``VIDEO_TS.*``, ``*.IFO``, ``*.BUP``) are ignored.

    :param video_ts_files: Sequence of files from a DVD'S "VIDEO_TS" directory
    """
    all_vobs = (
        filepath for filepath in video_ts_files
        if re.search(rf'{utils.fs.OS_SEP_REGEX}VIDEO_TS{utils.fs.OS_SEP_REGEX}VTS_\d+_\d+\.VOB$', filepath)
    )
    vob_sets = collections.defaultdict(lambda: [])
    for vob_path in all_vobs:
        # Find the largest sets of VOBs (e.g. VTS_01_*.VOB). We go by file size
        # instead of duration because we don't want to pick the VOB that is 20
        # hours of black (yes, that happened). The largest set should be the
        # main feature because it has the most complexity.
        vob_name = utils.fs.basename(vob_path)
        match = re.search(r'^(\w+_\d+)_\d+\.VOB$', vob_name)
        if match:
            set_name = match.group(1)
            vob_sets[set_name].append(vob_path)

    # Sort by VOB sets by combined size (largest first)
    sorted_vob_sets = sorted(
        vob_sets.values(),
        key=lambda vob_paths: sum(
            # file_size() returns `None` if size can't be determined
            (utils.fs.file_size(path) or 0)
            for path in vob_paths
        ),
        reverse=True,
    )

    # Return largest VOB set
    if sorted_vob_sets:
        return sorted_vob_sets[0]
    else:
        return []


def find_main_vob(video_ts_files):
    """
    Return largest VOB from the largest set of VOB files

    The main set of VOB files is determined by :func:`find_main_vobs`.

    If there are no VOB files, return `None`.

    :param video_ts_files: Sequence of files from a DVD'S "VIDEO_TS" directory
    """
    # Find largest VOB set (e.g. VTS_01_*.VOB)
    main_vobs = find_main_vobs(video_ts_files)
    if main_vobs:
        # Reverse sort VOBs by SIZE to get the largest VOB
        sorted_main_set = sorted(
            main_vobs,
            key=lambda vob_path: utils.fs.file_size(vob_path),
            reverse=True,
        )
        return sorted_main_set[0]


def vob2ifo(vob_path):
    """
    Translate VOB file to corresponding IFO file

    VOB and IFO files are used in VIDEO_TS directories on DVDs.

    :raise ValueError: if `vob_path` does not match expected naming pattern
    """
    filename = utils.fs.basename(vob_path)
    match = re.search(r'^(VTS_\d+_)\d+\.VOB$', filename)
    if match:
        return os.path.join(
            utils.fs.dirname(vob_path),
            match.group(1) + '0.IFO',
        )
    else:
        raise ValueError(f'This does not look like a VOB file: {vob_path}')


def for_each_video_ts(path):
    """
    Iterate over "VIDEO_TS" directories beneath `path`

    Subdirectories of `path` are searched recursively.
    """
    video_ts_paths = natsort.natsorted(
        glob.glob(
            f'{path}{utils.fs.OS_SEP_REGEX}**{utils.fs.OS_SEP_REGEX}VIDEO_TS',
            recursive=True,
        )
    )
    for video_ts_path in video_ts_paths:
        if os.path.isdir(video_ts_path):
            yield video_ts_path


@functools.lru_cache(maxsize=None)
def make_ffmpeg_input(path):
    """
    Make `path` palatable for ffmpeg

    - If path is a directory and contains a subdirectory named "BDMV", "bluray:"
      is prepended to `path`.

    - If path is a directory and contains a subdirectory named "VIDEO_TS", the
      largest .VOB file is returned. (ffmpeg does not support DVD directory
      structures.)

    - By default, `path` is returned unchanged.

    :param str path: Path to file or directory
    """
    # Detect Blu-ray
    if os.path.exists(os.path.join(path, 'BDMV')):
        return f'bluray:{path}'

    # Detect DVD
    if os.path.isdir(os.path.join(path, 'VIDEO_TS')):
        # FFmpeg doesn't seem to support reading DVD.  We work around this
        # by finding the first .VOB with a reasonable length.
        vobs = fs.file_list(os.path.join(path, 'VIDEO_TS'), extensions=('VOB',))
        _log.debug('All VOBs: %r', vobs)
        main_vobs = filter_main_videos(vobs)
        _log.debug('Main VOBs: %r', main_vobs)
        if main_vobs:
            return main_vobs[0]

    return str(path)



"""
Standardized configuration and setup howto
"""

from ... import __project_name__, utils


class Howto:
    def __init__(self, tracker_cls):
        self._tracker_cls = tracker_cls
        self._section = 0

    def _autobump(self, sections):
        self._section += len(sections)
        return '\n'.join(sections).strip()

    @property
    def current_section(self):
        return self._section

    @property
    def bump_section(self):
        self._section += 1
        return ''

    @property
    def introduction(self):
        return self._autobump((
            (
                f'{self._section}. How To Read This Howto\n'
                '\n'
                f'   {self._section}.1 Words in ALL_CAPS_AND_WITH_UNDERSCORES are placeholders.\n'
                f'   {self._section}.2 Everything after "$" is a terminal command.\n'
            ),
        ))

    @property
    def screenshots(self):
        return self._autobump((
            (
                f'{self._section}. Screenshots (Optional)\n'
                '\n'
                f'   {self._section}.1 Specify how many screenshots to make.\n'
                f'       $ {__project_name__} set trackers.{self._tracker_cls.name}.screenshots NUMBER_OF_SCREENSHOTS\n'
            ),
            (
                f'   {self._section}.2 Specify where to host images.\n'
                f'       $ {__project_name__} set trackers.{self._tracker_cls.name}.image_host IMAGE_HOST,...\n'
                f'       Supported services: ' + ', '.join(utils.imghosts.imghost_names()) + '\n'
                '\n'
                f'   {self._section}.3 Configure image hosting service.\n'
                f'       $ {__project_name__} upload-images IMAGE_HOST --help\n'
            ),
        ))

    @property
    def autoseed(self):
        return self._autobump((
            (
                f'{self._section}. Add Uploaded Torrents To Client (Optional)\n'
                '\n'
                f'   {self._section}.1 Specify which client to add uploaded torrents to.\n'
                f'       $ {__project_name__} set trackers.{self._tracker_cls.name}.add_to CLIENT_NAME\n'
                f'       Supported clients: ' + ', '.join(utils.btclient.client_names()) + '\n'
            ),
            (
                f'   {self._section}.2 Specify your client connection.\n'
                f'       $ {__project_name__} set clients.CLIENT_NAME.url URL\n'
                f'       $ {__project_name__} set clients.CLIENT_NAME.username USERNAME\n'
                f'       $ {__project_name__} set clients.CLIENT_NAME.password PASSWORD\n'
                '\n'
                f'{self._section + 1}. Copy Uploaded Torrents To Directory (Optional)\n'
                '\n'
                f'   $ {__project_name__} set trackers.{self._tracker_cls.name}.copy_to /path/to/directory\n'
            ),
        ))

    @property
    def upload(self):
        return self._autobump((
            (
                f'{self._section}. Upload\n'
                '\n'
                f'   $ {__project_name__} submit {self._tracker_cls.name} --help\n'
                f'   $ {__project_name__} submit {self._tracker_cls.name} /path/to/content\n'
            ),
        ))



"""
CLI argument types

All types return normalized values and raise ValueError for invalid values.

A custom error message can be provided by raising
:class:`argparse.ArgumentTypeError`.
"""

import argparse
import functools
import os

from . import types


def comma_separated(argtype):
    """
    Multiple comma-separated values

    :param argtype: Any callable that returns a validated object for one of the
        comma-separated values or raises :class:`ValueError`, :class:`TypeError`
        or :class:`argparse.ArgumentTypeError`

    :return: Sequence of `argtype` return values
    """

    def comma_separated(value):
        values = []
        for string in str(value).split(','):
            string = string.strip()
            if string:
                try:
                    values.append(argtype(string))
                except (ValueError, TypeError):
                    raise argparse.ArgumentTypeError(f'Invalid value: {string}')
        return values

    return comma_separated


def client(value):
    """Name of a supported BitTorrent client"""
    from . import btclient
    name = value.lower()
    if name in btclient.client_names():
        return name
    else:
        raise argparse.ArgumentTypeError(f'Unsupported client: {value}')


def content(value):
    """Existing path to release file(s)"""
    path = release(value)
    return existing_path(path)


def existing_path(value):
    """Path to existing path"""
    path = str(value)
    if not os.path.exists(path):
        raise argparse.ArgumentTypeError(f'No such file or directory: {value}')
    else:
        return path


def imghost(value):
    """Name of a image hosting service from :mod:`~.utils.imghosts`"""
    from . import imghosts
    if value in imghosts.imghost_names():
        return value.lower()
    else:
        raise argparse.ArgumentTypeError(f'Unsupported image hosting service: {value}')


def imghosts(value):
    """Comma-separated list of names of image hosting services from :mod:`~.utils.imghosts`"""
    names = []
    for name in value.split(','):
        name = name.strip()
        if name:
            names.append(imghost(name))
    return names


def integer(value):
    """Natural number (:class:`float` is rounded)"""
    try:
        return int(float(value))
    except (ValueError, TypeError):
        raise argparse.ArgumentTypeError(f'Not an integer: {value!r}')


def bool_or_none(value):
    """Convert `value` to :class:`~.types.Bool` or `None` if `value` is `None`"""
    if value is None:
        return None
    else:
        try:
            return types.Bool(value)
        except ValueError as e:
            raise argparse.ArgumentTypeError(e)


@functools.lru_cache(maxsize=None)
def number_of_screenshots(*, min, max):
    """
    Return function that returns how many screenshots to make or raises
    :class:`argparse.ArgumentTypeError`

    :param int min: Minimum number of screenshots
    :param int max: Maximum number of screenshots
    """

    def number_of_screenshots(value):
        """How many screenshots to make within allowed range"""
        try:
            return types.Integer(value, min=min, max=max)
        except ValueError as e:
            raise argparse.ArgumentTypeError(e)

    return number_of_screenshots


def option(value):
    """Name of a configuration option"""
    from .. import defaults
    if value in defaults.option_paths():
        return value.lower()
    else:
        raise argparse.ArgumentTypeError(f'Unknown option: {value}')


@functools.lru_cache(maxsize=None)
def one_of(values):
    """
    Return function that returns an item of `values` or raises
    :class:`argparse.ArgumentTypeError`

    :param values: Allowed values
    """
    values = tuple(values)

    def one_of_values(value):
        if value in values:
            return value
        else:
            raise argparse.ArgumentTypeError(f'Invalid value: {value}')

    return one_of_values


def regex(value):
    """:class:`re.Pattern` object"""
    import re
    try:
        return re.compile(str(value))
    except re.error as e:
        msg = f'Invalid regular expression: {value}: {e}'
        raise argparse.ArgumentTypeError(msg)


def release(value):
    """Same as :func:`content`, but doesn't have to exist"""
    from .. import errors
    from . import predbs
    path = str(value)
    try:
        predbs.assert_not_abbreviated_filename(path)
    except errors.SceneAbbreviatedFilenameError as e:
        raise argparse.ArgumentTypeError(e)
    else:
        return path


def predb_name(value):
    """Name of a scene release database from :mod:`~.utils.predbs`"""
    from . import predbs
    if value in predbs.predb_names():
        return value.lower()
    else:
        raise argparse.ArgumentTypeError(f'Unsupported scene release database: {value}')


def predb(value):
    """
    :class:`~.PredbApiBase` instance from a corresponding
    :attr:`~.PredbApiBase.name`
    """
    from . import predbs
    try:
        return predbs.predb(value.lower())
    except ValueError as e:
        raise argparse.ArgumentTypeError(e)


def timestamp(value):
    """See :func:`.timestamp.parse`"""
    from . import timestamp
    try:
        return timestamp.parse(value)
    except (ValueError, TypeError) as e:
        raise argparse.ArgumentTypeError(e)


def tracker(value):
    """Name of a tracker from :mod:`~.trackers`"""
    from .. import trackers
    if value in trackers.tracker_names():
        return value.lower()
    else:
        raise argparse.ArgumentTypeError(f'Unsupported tracker: {value}')


def webdb(value):
    """Name of a movie/series database from :mod:`~.webdbs`"""
    from . import webdbs
    if value in webdbs.webdb_names():
        return value.lower()
    else:
        raise argparse.ArgumentTypeError(f'Unsupported database: {value}')



"""
Find, download and re-upload poster for movie, series or season
"""

import functools

from .... import jobs, utils
from .base import CommandBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class poster(CommandBase):
    """Download, resize and re-upload poster from IMDb or similar website"""

    names = ('poster',)

    argument_definitions = {
        'RELEASE': {
            'type': utils.argtypes.release,
            'help': 'Release name or path to release content',
        },
        ('--db', '-d'): {
            'type': utils.argtypes.webdb,
            'help': ('Case-insensitive database name (default: imdb)\n'
                     'Supported databases: ' + ', '.join(utils.webdbs.webdb_names())),
            'default': None,
        },
        ('--upload-to', '-u'): {
            'type': utils.argtypes.imghost,
            'metavar': 'IMAGE_HOST',
            'help': ('Case-insensitive name of image hosting service\n'
                     'Supported services: ' + ', '.join(utils.imghosts.imghost_names())),
            'default': None,
        },
        ('--width', '-w'): {
            'help': 'Maximum poster width in pixels (default: 500)',
            'type': utils.argtypes.integer,
            'default': 500,
        },
        ('--height', '-t'): {
            'help': 'Maximum poster height in pixels (default: 1000)',
            'type': utils.argtypes.integer,
            'default': 1000,
        },
    }

    @functools.cached_property
    def release_name(self):
        return utils.release.ReleaseName(self.args.RELEASE)

    @functools.cached_property
    def webdb(self):
        if self.args.db:
            return utils.webdbs.webdb(self.args.db)
        elif self.release_name.type in (utils.release.ReleaseType.season,
                                        utils.release.ReleaseType.episode):
            return utils.webdbs.webdb('tvmaze')
        else:
            return utils.webdbs.webdb('imdb')

    @functools.cached_property
    def imghost(self):
        if self.args.upload_to:
            return utils.imghosts.imghost(
                name=self.args.upload_to,
                options=self.config['imghosts'][self.args.upload_to],
                cache_directory=self.cache_directory,
            )

    @functools.cached_property
    def jobs(self):
        return (
            self.webdb_job,
            self.poster_job,
        )

    @functools.cached_property
    def webdb_job(self):
        return jobs.webdb.WebDbSearchJob(
            home_directory=self.cache_directory,
            cache_directory=self.cache_directory,
            ignore_cache=self.args.ignore_cache,
            query=utils.webdbs.Query.from_release(self.release_name),
            db=self.webdb,
            callbacks={
                'output': self.start_poster_job,
            },
        )

    def start_poster_job(self, webdb_id):
        self.poster_job.start(
            webdb_id=webdb_id,
            season=self.release_name.only_season,
        )

    @functools.cached_property
    def poster_job(self):
        return jobs.poster.PosterJob(
            home_directory=self.home_directory if self.imghost else '.',
            cache_directory=self.cache_directory,
            ignore_cache=self.args.ignore_cache,
            autostart=False,
            webdb_id=None,
            webdb=self.webdb,
            imghost=self.imghost,
            width=self.args.width,
            height=self.args.height,
        )
